"""Integrated fluxes."""

from iris import Constraint
from iris.analysis import SUM
from iris.cube import CubeList
from iris.util import guess_coord_axis
import numpy as np

from ..const import get_planet_radius
from ..coord import nearest_coord_value, vertical_cross_section_area
from ..exceptions import _warn
from ..model import um

__all__ = (
    "horizontal_fluxes_through_region_boundaries",
    "net_horizontal_flux_to_region",
)


def horizontal_fluxes_through_region_boundaries(
    scalar_cube,
    region,
    u,
    v,
    r_planet=None,
    vertical_constraint=None,
    warn_thresh=10,
    model=um,
):
    """Calculate horizontal fluxes through planes of a rectangular region."""
    perpendicular_wind_cmpnts = {um.x: u, um.y: v}

    if r_planet is None:
        r = get_planet_radius(scalar_cube)
    else:
        r = r_planet

    total_h_fluxes = CubeList()
    for bound in region:
        this_coord = bound["coord"]
        (
            other_coord,
            other_min,
            other_max,
        ) = region._perpendicular_side_limits(bound["name"])
        nearest = nearest_coord_value(scalar_cube, this_coord, bound["value"])
        if abs(nearest - bound["value"]) >= warn_thresh:
            _warn(
                f"Nearest value is {np.round(nearest - bound['value'], 2)}"
                f" deg away from the given value of {this_coord}",
            )
        vcross_cnstr = Constraint(**{this_coord: nearest})
        vcross_cnstr &= vertical_constraint

        if other_max >= other_min:
            vcross_cnstr &= Constraint(
                **{other_coord: lambda x: other_min <= x <= other_max}
            )
            cube = scalar_cube.extract(vcross_cnstr)
        else:
            vcross_cnstr &= Constraint(
                **{other_coord: lambda x: (other_max >= x) or (other_min <= x)}
            )
            cube = scalar_cube.extract(vcross_cnstr)
        cube_slice = next(
            cube.slices([cube.coord(axis="z").name(), other_coord])
        )
        vcross_area = vertical_cross_section_area(cube_slice, r_planet=r)

        # Calculate side flux (2d)
        cube = (
            perpendicular_wind_cmpnts[this_coord].extract(vcross_cnstr)
            * cube
            * vcross_area
        )
        cube.rename(
            f"{scalar_cube.name()}_flux_through_{bound['name']}_boundary"
        )
        # Total flux
        collapsible_dims = [
            i
            for i in cube.dim_coords
            if guess_coord_axis(i) in ["Z", "Y", "X"]
        ]
        cube_total = cube.collapsed(collapsible_dims, SUM)
        total_h_fluxes.append(cube_total)
    return total_h_fluxes


def net_horizontal_flux_to_region(
    scalar_cube,
    region,
    u,
    v,
    r_planet=None,
    vertical_constraint=None,
    model=um,
):
    """Calculate horizontal fluxes and add them to get the net result."""
    total_h_fluxes = horizontal_fluxes_through_region_boundaries(
        scalar_cube,
        region,
        u,
        v,
        r_planet=r_planet,
        vertical_constraint=vertical_constraint,
        model=model,
    )
    net_flux = (
        total_h_fluxes.extract_cube(
            Constraint(cube_func=lambda x: "through_west" in x.name())
        )
        - total_h_fluxes.extract_cube(
            Constraint(cube_func=lambda x: "through_east" in x.name())
        )
        + total_h_fluxes.extract_cube(
            Constraint(cube_func=lambda x: "through_south" in x.name())
        )
        - total_h_fluxes.extract_cube(
            Constraint(cube_func=lambda x: "through_north" in x.name())
        )
    )
    net_flux.rename(f"net_{scalar_cube.name()}_horizontal_flux_to_region")
    net_flux.attributes["region_str"] = str(region)

    return net_flux



"""Utilities for working with LFRic output."""

from collections.abc import Sequence
from pathlib import Path
from typing import Callable, Optional, Union

import iris
import iris.coords
from iris.cube import Cube, CubeList
import iris.exceptions
from iris.experimental.ugrid import PARSE_UGRID_ON_LOAD, MeshCoord, load_mesh
from iris.experimental.ugrid.mesh import Mesh
import iris.util
import numpy as np

from .io import load_vert_lev
from .model import lfric
from .model.base import Model

__all__ = (
    "add_um_height_coord",
    "add_equally_spaced_height_coord",
    "clean_attrs",
    "fix_time_coord",
    "load_lfric_mesh",
    "load_lfric_raw",
    "replace_level_coord_with_height",
    "replace_mesh",
    "simple_regrid_lfric",
    "ugrid_spatial",
    "ugrid_spatial_mean",
)


def add_um_height_coord(
    cube: Cube, field: str, filename: str, path_to_levels_file: Path
) -> Cube:
    """
    Callback for `iris.load` specifically for LFRic data with vertical coord.

    Adds a vertical coordinate of level heights from a UM vertical levels file.

    Parameters
    ----------
    path_to_levels_file: Path-like
        Path to the vertical levels file.

    Examples
    --------
    >>> from functools import partial
    >>> iris.load(
        filename,
        callback=partial(
            add_um_height_coord,
            path_to_levels_file=Path('/path/to/vertlevs_L38_29t_9s_40km')
        )
    )
    """
    lev_coords = [
        i.name() for i in cube.dim_coords if i.name().endswith("_levels")
    ]
    if len(lev_coords) == 1:
        thlev = load_vert_lev(path_to_levels_file, lev_type="theta")
        rholev = load_vert_lev(path_to_levels_file, lev_type="rho")
        lev_coord = cube.coord(lev_coords[0])
        if "full" in lev_coord.name().lower():
            points = thlev
        else:
            points = rholev
        hgt_coord = iris.coords.AuxCoord(
            points,
            long_name="level_height",
            units="m",
            attributes={"positive": "up"},
        )
        hgt_coord.guess_bounds()
        cube.add_aux_coord(hgt_coord, data_dims=cube.coord_dims(lev_coord))

    return cube


def add_equally_spaced_height_coord(
    cube: Cube, field: str, filename: str, model_top_height: float
) -> Cube:
    """
    Callback for `iris.load` specifically for LFRic data with vertical coord.

    Adds a vertical coordinate of level heights assuming equally spaced levels.

    Examples
    --------
    >>> from functools import partial
    >>> iris.load(
        filename,
        callback=partial(
            add_equally_spaced_height_coord,
            model_top_height=32000
        )
    )
    """
    lev_coords = [
        i.name() for i in cube.dim_coords if i.name().endswith("_levels")
    ]
    if len(lev_coords) == 1:
        lev_coord = cube.coord(lev_coords[0])
        if "full" in lev_coord.name().lower():
            points = np.linspace(0, model_top_height, lev_coord.shape[0])
        else:
            points = np.linspace(0, model_top_height, lev_coord.shape[0] + 1)
            points = 0.5 * (points[:-1] + points[1:])
        hgt_coord = iris.coords.AuxCoord(
            points,
            long_name="level_height",
            units="m",
            attributes={"positive": "up"},
        )
        hgt_coord.guess_bounds()
        cube.add_aux_coord(hgt_coord, data_dims=cube.coord_dims(lev_coord))

    return cube


def clean_attrs(cube: Cube, field: str, filename: str) -> Cube:
    """
    Callback for `iris.load` to clean up trivial attributes in LFRic data.

    Needed for concatenating cubes.
    """
    for attr in ["timeStamp", "uuid"]:
        try:
            cube.attributes.pop(attr)
        except KeyError:
            pass
    return cube


def fix_time_coord(
    cube: Cube,
    field: str,
    filename: str,
    downgrade_to_scalar: Optional[bool] = False,
) -> Cube:
    """
    Callback function for `iris.load` specifically for UGRID data.

    All field variables should have time as a coord, and mesh variables
    will not have a time coord. So ignore anything that doesn't have a
    time coord. Note that mesh info will still be loaded in the
    coordinate part of the cube.

    Make sure that time coordinate is correct type, seems to always be
    loaded as a basic AuxCoord rather than a Scalar or DimCoord.
    """
    # Ignore cubes with no time coordinate
    if len(cube.coords("time")) == 0:
        raise iris.exceptions.IgnoreCubeException

    # Else, fix metadata in variables
    else:
        tcoord = cube.coord("time")
        # If only 1 time coordinate value, downgrade AuxCoord to Scalar
        if downgrade_to_scalar and tcoord.points.size == 1:
            cube = cube.extract(iris.Constraint(time=tcoord.cell(0)))
        # Else, upgrade AuxCoord to DimCoord
        else:
            iris.util.promote_aux_coord_to_dim_coord(cube, tcoord)
    return cube


def load_lfric_mesh(
    fname: str = "mesh.nc", var_name: str = "dynamics"
) -> Mesh:
    """Load LFRic mesh from a netCDF file."""
    with PARSE_UGRID_ON_LOAD.context():
        loaded_mesh = load_mesh(fname, var_name=var_name)

    return loaded_mesh


def load_lfric_raw(
    fnames: Sequence[Union[Path, str]],
    callback: Optional[Callable] = None,
    drop_coord: Optional[Sequence[str]] = (),
) -> CubeList:
    """Load raw LFRic data."""
    with PARSE_UGRID_ON_LOAD.context():
        cl_raw = iris.load(fnames, callback=callback)
    cl_raw = CubeList(clean_attrs(cube, None, None) for cube in cl_raw)
    for coord in drop_coord:
        for cube in cl_raw:
            try:
                cube.remove_coord(coord)
            except iris.exceptions.CoordinateNotFoundError:
                pass
    cl_raw = cl_raw.concatenate(check_aux_coords=False)
    return cl_raw


def replace_level_coord_with_height(cube: Cube) -> Cube:
    """
    Remove full_ or half_levels coordinate and replace it with level_height.

    Assumes that the cube contains level_height already.

    See Also
    --------
    add_equally_spaced_height_coord
    """
    lev_coords = [
        i.name() for i in cube.dim_coords if i.name().endswith("_levels")
    ]
    if len(lev_coords) == 1:
        lev_coord = cube.coord(lev_coords[0])
        cube.remove_coord(lev_coord)
        iris.util.promote_aux_coord_to_dim_coord(cube, "level_height")
    return cube


def replace_mesh(cube: Cube, new_mesh: Mesh) -> Cube:
    """Replace mesh in a 1d cube by creating a new copy of that cube."""
    mesh_x = MeshCoord(mesh=new_mesh, location="face", axis="x")
    mesh_y = MeshCoord(mesh=new_mesh, location="face", axis="y")

    new_cube = Cube(cube.data, aux_coords_and_dims=[(mesh_x, 0), (mesh_y, 0)])
    new_cube.metadata = cube.metadata
    return new_cube


def _regrid_cubes_on_edges(
    src_cube: Cube, tgt_cube: Cube, method: Optional[str] = "auto"
) -> Cube:
    """Temporary solution to regrid cubes defined on cell edges."""
    # Using the approach in gungho/rose-stem/app/plot/bin/plot_zonal.py
    from iris._concatenate import _CubeSignature
    from iris.analysis.cartography import get_xy_grids
    from scipy.interpolate import griddata

    src_x = src_cube.coord("longitude").points
    src_y = src_cube.coord("latitude").points
    tgt_x2d, tgt_y2d = get_xy_grids(tgt_cube)

    non_xy_coords = [
        name
        for c in src_cube.dim_coords
        if (name := c.name()) not in ["longitude", "latitude"]
    ]

    result = CubeList()
    for src_2d_slice in src_cube.slices_over(non_xy_coords):
        # Interpolate using delaunay triangularization
        if method == "auto":
            reg_data_linear = griddata(
                (src_x, src_y),
                src_2d_slice.data,
                (tgt_x2d, tgt_y2d),
                method="linear",
            )
            reg_data_nearest = griddata(
                (src_x, src_y),
                src_2d_slice.data,
                (tgt_x2d, tgt_y2d),
                method="nearest",
            )
            reg_data_linear[np.isnan(reg_data_linear)] = reg_data_nearest[
                np.isnan(reg_data_linear)
            ]
            reg_data = reg_data_linear
        else:
            reg_data = griddata(
                (src_x, src_y),
                src_2d_slice.data,
                (tgt_x2d, tgt_y2d),
                method=method,
            )
        reg_cube_slice = tgt_cube.copy(data=reg_data)
        for scalar_coord in _CubeSignature(src_2d_slice).scalar_coords:
            reg_cube_slice.add_aux_coord(scalar_coord)
        result.append(reg_cube_slice)
    result = result.merge_cube()
    dim_order = [*range(src_cube.ndim + 1)]
    dim_order[0], dim_order[1] = 1, 0
    result.transpose(dim_order)
    result.metadata = src_cube.metadata
    return result


def simple_regrid_lfric(
    cube_list: CubeList,
    tgt_cube: Cube,
    ref_cube_constr: Optional[str] = "air_potential_temperature",
    interp_vertically: Optional[bool] = True,
    model: Optional[Model] = lfric,
) -> CubeList:
    """Quick&dirty regridding of LFRic data to a common height/lat/lon grid."""
    from esmf_regrid.experimental.unstructured_scheme import (
        MeshToGridESMFRegridder,
    )

    # Horizontal regridding
    result_h = CubeList()
    ref_cube = cube_list.extract_cube(ref_cube_constr)
    regridder = MeshToGridESMFRegridder(
        ref_cube, tgt_cube, method="conservative"
    )
    for cube in cube_list:
        if cube.location == "edge":
            cube_reg = _regrid_cubes_on_edges(cube, tgt_cube)
        else:
            cube_reg = regridder(cube)
        result_h.append(cube_reg)
    result = result_h
    if interp_vertically:
        # Vertical interpolation
        result_v = iris.cube.CubeList()
        tgt_points = (model.z, ref_cube.coord(model.z).points)
        for cube in result:
            if len(
                [
                    i.name()
                    for i in cube.dim_coords
                    if i.name().endswith("_levels")
                ]
            ):
                result_v.append(
                    replace_level_coord_with_height(cube).interpolate(
                        [tgt_points], iris.analysis.Linear()
                    )
                )
            else:
                result_v.append(cube)
        result = result_v
    return result


def ugrid_spatial(
    cube: Cube,
    aggr: str,
    model: Optional[Model] = lfric,
    **kwargs,
) -> Cube:
    """Collapse a UGRID `iris.cube.Cube` over x and y coord with no weights."""
    cube_copy = cube.copy()
    tmp_coord = iris.coords.AuxCoord(
        points=np.arange(cube.coord(model.x).shape[0]),
        long_name="mesh_coordinates",
    )

    cube_copy.add_aux_coord(tmp_coord, data_dims=cube.coord_dims(model.x))

    cube_copy.remove_coord(model.x)
    cube_copy.remove_coord(model.y)

    cube_aggr = cube_copy.collapsed(
        tmp_coord.name(), getattr(iris.analysis, aggr.upper()), **kwargs
    )
    cube_aggr.remove_coord(tmp_coord)
    return cube_aggr


def ugrid_spatial_mean(
    cube: Cube, model: Optional[Model] = lfric, **kwargs
) -> Cube:
    """Average a UGRID `iris.cube.Cube` spatially."""
    return ugrid_spatial(cube, "mean", model=model, **kwargs)



"""Command line interface (CLI) support for the aeolus entry-point."""

import click
from click_default_group import DefaultGroup

from ._version import version as __version__

__all__ = ["main"]

CONTEXT_SETTINGS = {"help_option_names": ["-h", "--help"]}


@click.group(
    cls=DefaultGroup,
    default="pp",
    default_if_no_args=True,
    invoke_without_command=True,
    context_settings=CONTEXT_SETTINGS,
)
@click.option(
    "-v",
    "--version",
    is_flag=True,
    help="Show the package version.",
)
def main(version: bool) -> None:
    """To get help for other commands, simply use "aeolus COMMAND --help"."""
    if version:
        click.echo("version ", nl=False)
        click.secho(f"{__version__}")


@main.command(no_args_is_help=True)
@click.option(
    "-m",
    "--model",
    default="lfric",
    type=click.Choice(["lfric", "um"]),
    help="Model",
)
@click.option("-i", "--inpdir", type=str, help="Input directory")
@click.option("-o", "--outdir", type=str, help="Output directory")
@click.option(
    "-l", "--label", type=str, required=True, help="Simulation label"
)
@click.option(
    "-p", "--planet", type=str, required=True, help="Planet configuration"
)
@click.option(
    "-c", "--c_num", type=str, default="C48", help="Cubed Sphere Mesh Number"
)
@click.option(
    "--ref_cube",
    type=str,
    default="air_potential_temperature",
    help="Reference cube, to which coordinates all data will be regridded",
)
@click.option(
    "--level_height",
    type=str,
    default="uniform",
    help="Type of the vertical level height coordinate",
)
@click.option(
    "--top_height",
    type=float,
    default=40e3,
    help="If level_height=uniform, set the model top height.",
)
@click.option(
    "--time_prof",
    default="inst",
    help="Type of the time output",
    show_default=True,
    type=click.Choice(["inst", "mean"]),
)
@click.option(
    "--file_chunk_size",
    type=int,
    default=1000,
    help="Number of files to process in one go.",
)
@click.option(
    "--nlat",
    type=int,
    default=90,
    help="Number of latitudes in the target grid.",
)
@click.option(
    "--nlon",
    type=int,
    default=144,
    help="Number of longitudes in the target grid.",
)
def pp(
    model: str,
    inpdir: str,
    outdir: str,
    label: str,
    planet: str,
    c_num: str,
    ref_cube: str,
    level_height: str,
    top_height: float,
    time_prof: str,
    file_chunk_size: int,
    nlat: int,
    nlon: int,
) -> None:
    """Post-process model output."""
    from .postprocess import process_lfric  # noqa

    click.secho(f"Processing {model} output...")
    if model == "lfric":
        fname_out = process_lfric(
            inpdir,
            outdir,
            label,
            planet,
            c_num,
            ref_cube,
            level_height,
            top_height,
            time_prof,
            file_chunk_size,
            nlat,
            nlon,
        )
        click.secho(f"Saved to {fname_out}")



"""Interface to metpy calc functions."""

import functools

import cf_units
from iris.cube import Cube
import metpy.units as metunits
import numpy as np
import pint
import xarray as xr

from ..exceptions import UnitFormatError

__all__ = ("preprocess_iris",)


def preprocess_iris(f):
    """
    Wrap a function from `metpy.calc` for it to accept iris cubes as arguments.

    In addition, this decorator converts `metpy.calc` output by using the first
    input argument as a 'donor' cube.
    Note this works only for functions that preserve dimensions and may not
    work with some units. Now that metpy has xarray preprocessor,
    this decorator depends on it.
    """

    # Define local functions
    def to_xarray(cube):
        """Convert `iris cube` to `DataArray` and format units correctly."""
        _unit = None
        for ut_format in set(cf_units.UT_FORMATS):
            try:
                _unit = metunits.units(cube.units.format(ut_format))
            except pint.errors.DimensionalityError:
                pass
        if _unit is None:
            raise UnitFormatError(
                f"Unable to convert \n{repr(cube)} units \nto metpy units"
            )
        arr = xr.DataArray.from_iris(cube)
        arr.attrs["units"] = str(_unit)
        return arr

    def to_iris(donor_cube, arr, name):
        """Convert metpy calc result to `iris.cube.Cube`."""
        try:
            data = arr.magnitude
            units = str(arr.units).replace(" ** ", "^").replace(" * ", " ")
        except AttributeError:
            # output is probably a numpy array without units
            data = np.asarray(arr)
            units = "1"
        cube_out = donor_cube.copy(data=data)
        try:
            cube_out.units = units
        except ValueError:
            cube_out.units = "unknown"
        cube_out.rename(name)
        cube_out.attributes.pop("STASH", None)
        return cube_out

    @functools.wraps(f)
    def wrapper(*args, **kwargs):
        nargs = []
        _cube = None  # a donor cube with metadata

        # Loop over cubes and re-format units
        for arg in args:
            if isinstance(arg, Cube):
                if arg.ndim > 0:
                    # TODO: make this flexible
                    _cube = arg
                elif _cube is None:
                    _cube = arg

                nargs.append(to_xarray(arg))
            else:
                nargs.append(arg)

        kwargs = {
            k: (to_xarray(v) if isinstance(v, Cube) else v)
            for k, v in kwargs.items()
        }

        # Call the decorated function
        out = f(*nargs, **kwargs)
        if _cube is None:
            # Inputs are not iris cubes
            return out

        # Convert ouput to iris
        # TODO: check masked array behaviour
        if isinstance(out, (tuple, list, set)):
            res = []
            for i, iout in enumerate(out):
                res.append(to_iris(_cube, iout, f"{f.__name__}_output_{i}"))
            res = tuple(res)
        else:
            res = to_iris(_cube, out, f.__name__)
        return res

    return wrapper



"""Functions for calculating synthetic observations."""

import warnings

from iris.analysis import SUM
from iris.coords import AuxCoord, DimCoord
from iris.cube import Cube
from iris.exceptions import CoordinateNotFoundError as CoNotFound
from iris.util import reverse
import numpy as np

from .coord import roll_cube_pm180
from .model import um

__all__ = (
    "calc_geom_mean_mirrored",
    "calc_stellar_flux",
    "calc_transmission_spectrum",
    "calc_transmission_spectrum_day_night_average",
    "read_normalized_stellar_flux",
    "read_spectral_bands",
)


def calc_geom_mean_mirrored(cube_a, cube_b, add_shift=0, model=um):
    """
    Calculate geometric mean of 2 cubes, one of them flipped along the x-axis.

    This function can be used to get an average transmission flux
    calculated separately from the day- and night-side perspective.

    cube_a: iris.cube.Cube
        Cube with an x-coordinate.
    cube_b: iris.cube.Cube
        Another cube, to be flipped before being averaged with cube A.
    add_shift: int
        Additional shift for the data along the x-coordinate.
    model: aeolus.model.Model, optional
        Model class with relevant variable names.
    """
    # Get x-coordinate from the 1st cube.
    x_coord = cube_a.coord(model.x)

    # Roll and reverse the 2nd cube by 180 degrees
    # This is specific to the UM output
    cube_b_flipped = reverse(
        roll_cube_pm180(cube_b, add_shift=add_shift, model=model), model.x
    )
    cube_b_flipped.replace_coord(x_coord)

    # Calculate geometric mean of the two cubes
    geom_mean = (cube_a * cube_b_flipped) ** 0.5
    return geom_mean


def calc_stellar_flux(spectral_file, stellar_constant_at_1_au):
    """
    Calculate the stellar flux per spectral band.

    Parameters
    ----------
    spectral_file: pathlib.Path
        Path to the location of a SOCRATES spectral file.
    stellar_constant_at_1_au: float or iris.cube.Cube
        Stellar constant at 1 AU [W m-2].

    Returns
    -------
    iris.cube.Cube
        Stellar flux per spectral band [W m-2].
    """
    # Ensure that an input constant is an iris cube
    if not isinstance(stellar_constant_at_1_au, Cube):
        stellar_constant_at_1_au = Cube(
            stellar_constant_at_1_au,
            long_name="stellar_constant_at_1_au",
            units="W m-2",
        )
    normalized_stellar_flux = read_normalized_stellar_flux(spectral_file)
    stellar_flux = normalized_stellar_flux * stellar_constant_at_1_au
    stellar_flux.rename("stellar_flux")
    return stellar_flux


def calc_transmission_spectrum(
    trans_flux,
    spectral_file=None,
    stellar_constant_at_1_au=None,
    stellar_radius=None,
    planet_top_of_atmosphere=None,
    model=um,
):
    r"""
    Convert the transmission flux to a planetary-stellar radius ratio.

    Parameters
    ----------
    trans_flux: iris.cube.Cube
        Transmission flux on spectral bands and optionally lats and lons.
        In the Met Office Unified Model this is
        STASH items 555, 556, 755, 756 in section 1.
    spectral_file: pathlib.Path
        Path to the location of a SOCRATES spectral file.
    stellar_constant_at_1_au: float or iris.cube.Cube
        Stellar constant at 1 AU [W m-2].
    stellar_radius: float or iris.cube.Cube
        Stellar radius [m].
    planet_top_of_atmosphere: float or iris.cube.Cube
        The extent of the planetary atmosphere [m].
    model: aeolus.model.Model, optional
        Model class with relevant variable names.

    Returns
    -------
    iris.cube.Cube
        The ratio of the effective planetary radius to the stellar radius
        per spectral band [1].
        Spectral band centres [m] is attached as an auxiliary coordinate.

    Notes
    -----
    The transmission spectrum is the ratio of the effective planetary
    radius to the stellar radius calculated per spectral band:

    .. math::

        \frac{R_p (\nu)}{R_s} = \sqrt{(\frac{R_{p,TOA}}{R_s})^2 -
         \frac{\sum_{lat,lon}^{}F_{trans} (\nu)}{F_{stellar} (\nu)}}

    where
    :math:`R_p(\nu)` is the effective planetary radius,
    :math:`R_s` is the stellar radius,
    :math:`R_{p,TOA}` is the extent of the planetary atmosphere
    (i.e. the sum of the planetary radius and the height of the model domain),
    :math:`\sum_{lat,lon}^{}F_{trans}(\nu)` is the total transmitted flux,
    :math:`F_{stellar}(\nu)` is the stellar flux.

    Smaller gas abundance leads to the corresponding limb to be
    "more transmissive", which leads it having smaller transit depth.
    """
    # Ensure that input constants are iris cubes
    if not isinstance(stellar_constant_at_1_au, Cube):
        stellar_constant_at_1_au = Cube(
            stellar_constant_at_1_au,
            long_name="stellar_constant_at_1_au",
            units="W m-2",
        )
    if not isinstance(stellar_radius, Cube):
        stellar_radius = Cube(
            stellar_radius,
            long_name="stellar_radius",
            units="m",
        )
    if not isinstance(planet_top_of_atmosphere, Cube):
        planet_top_of_atmosphere = Cube(
            planet_top_of_atmosphere,
            long_name="planet_top_of_atmosphere",
            units="m",
        )
    # Calculate stellar flux
    stellar_flux = calc_stellar_flux(spectral_file, stellar_constant_at_1_au)

    # Sum transmission flux over all latitudes and longitudes
    try:
        with warnings.catch_warnings():
            warnings.filterwarnings("ignore", category=UserWarning)
            trans_flux = trans_flux.collapsed([model.y, model.x], SUM)
    except CoNotFound:
        pass
    # trans_flux.rename("shortwave_transmission_flux")
    trans_flux.units = "W m-2"

    # Calculate the ratio of the total transmitted flux to the stellar flux
    for coord in trans_flux.dim_coords:
        if coord.name().lower() in ["pseudo", "pseudo_level"]:
            coord.rename("spectral_band_index")
            break
    flux_ratio = trans_flux / stellar_flux

    # The ratio of the effective planetary radius to the stellar radius
    rp_eff_over_rs_squared = (
        planet_top_of_atmosphere / stellar_radius
    ) ** 2 - flux_ratio
    # Make all negative values zero
    rp_eff_over_rs_squared = rp_eff_over_rs_squared.copy(
        data=rp_eff_over_rs_squared.core_data().clip(min=0.0)
    )
    rp_eff_over_rs = rp_eff_over_rs_squared ** (0.5)
    rp_eff_over_rs.rename(
        "ratio_of_effective_planetary_radius_to_stellar_radius"
    )

    # Find spectral band centers
    spectral_bands = read_spectral_bands(spectral_file)
    spectral_band_centres = 0.5 * (
        spectral_bands["lower_wavelength_limit"]
        + spectral_bands["upper_wavelength_limit"]
    )
    spectral_bands_coord = AuxCoord(
        spectral_band_centres,
        long_name="spectral_band_centres",
        units="m",
    )
    spectral_bands_ll = AuxCoord(
        spectral_bands["lower_wavelength_limit"],
        long_name="spectral_band_lower_limit",
        units="m",
    )
    spectral_bands_ul = AuxCoord(
        spectral_bands["upper_wavelength_limit"],
        long_name="spectral_band_upper_limit",
        units="m",
    )

    # Attach spectral bands to the resulting cube as an auxiliary coordinate
    coord_dim = rp_eff_over_rs.coord_dims("spectral_band_index")[0]
    rp_eff_over_rs.add_aux_coord(spectral_bands_coord, data_dims=(coord_dim,))
    rp_eff_over_rs.add_aux_coord(spectral_bands_ll, data_dims=(coord_dim,))
    rp_eff_over_rs.add_aux_coord(spectral_bands_ul, data_dims=(coord_dim,))

    return rp_eff_over_rs


def calc_transmission_spectrum_day_night_average(
    trans_flux_day,
    trans_flux_night,
    add_shift=0,
    spectral_file=None,
    stellar_constant_at_1_au=None,
    stellar_radius=None,
    planet_top_of_atmosphere=None,
    model=um,
):
    r"""
    Convert the transmission flux to a planetary-stellar radius ratio.

    For UM output, this function averages the flux calculated from the day-side
    and the night-side of the planet.
    Why does it use a geometric mean? The reason to use a geometric average
    instead of an arithmetic average is that the optical depths are added.
    The flux decreases via Beer's law (i.e., it's proportional to
    :math:`exp[-optical depth]`) so when you multiply the dayside fluxes and
    nightside fluxes together and take a square root, you end up with
    :math:`exp[-( optical depth 1 + optical depth 2)/2]`. Since each optical
    depth is double the optical depth for it's respective side, the factors of
    two cancel and you end up with :math:`exp[-(true optical depth)]`.

    Parameters
    ----------
    trans_flux_day: iris.cube.Cube
        Transmission flux on spectral bands and optionally lats and lons.
        Day-side perspective.
        In the Met Office Unified Model this is
        STASH items 555, 556, 755, 756 in section 1.
    trans_flux_night: iris.cube.Cube
        Samea as day, but for the night-side calculation.
    add_shift: int, optional
        Additional shift of data along the x-coordinate.

    See Also
    --------
    aeolus.synthobs.calc_transmission_spectrum
    """
    # Average the day and night flux
    trans_flux = calc_geom_mean_mirrored(
        trans_flux_day, trans_flux_night, add_shift=add_shift, model=model
    )
    return calc_transmission_spectrum(
        trans_flux,
        spectral_file=spectral_file,
        stellar_constant_at_1_au=stellar_constant_at_1_au,
        stellar_radius=stellar_radius,
        planet_top_of_atmosphere=planet_top_of_atmosphere,
        model=model,
    )


def read_normalized_stellar_flux(spectral_file):
    """
    Read the normalized stellar flux per band from a SOCRATES spectral file.

    Parameters
    ----------
    spectral_file: pathlib.Path
        Path to the location of the SOCRATES spectral file.

    Returns
    -------
    iris.cube.Cube
        Normalized stellar flux per spectral band [1].
    """
    with spectral_file.open("r") as f:
        lines = []
        band_block = False
        for line in f:
            if line.startswith("*BLOCK: TYPE =    2"):
                band_block = True
            if band_block:
                if line.startswith("*END"):
                    break
                elif line.lstrip().split(" ")[0].isnumeric():
                    lines.append(
                        tuple(
                            [
                                float(i)
                                for i in line.lstrip().rstrip("\n").split(" ")
                                if i
                            ]
                        )
                    )
    normalized_stellar_flux_arr = np.array(
        lines,
        dtype=[
            ("spectral_band_index", "u4"),
            ("normalized_stellar_flux", "f4"),
        ],
    )
    # Compose an iris cube
    spectral_band_index = DimCoord(
        normalized_stellar_flux_arr["spectral_band_index"],
        long_name="spectral_band_index",
        units="1",
    )
    normalized_stellar_flux = Cube(
        normalized_stellar_flux_arr["normalized_stellar_flux"],
        long_name="normalized_stellar_flux",
        dim_coords_and_dims=[(spectral_band_index, 0)],
        units="1",
    )
    return normalized_stellar_flux


def read_spectral_bands(spectral_file):
    """
    Read spectral bands from a SOCRATES spectral file.

    Parameters
    ----------
    spectral_file: pathlib.Path
        Path to the location of a SOCRATES spectral file.

    Returns
    -------
    numpy.ndarray
        An array with a list of tuples describing spectral bands.
        Tuple structure:
        (spectral_band_index, lower_wavelength_limit, upper_wavelength_limit).
        Units [m] as stated in a spectral file but not checked directly.
    """
    with spectral_file.open("r") as f:
        lines = []
        band_block = False
        for line in f:
            if line.startswith("*BLOCK: TYPE =    1"):
                band_block = True
            if band_block:
                if line.startswith("*END"):
                    break
                elif line.lstrip().split(" ")[0].isnumeric():
                    lines.append(
                        tuple(
                            [
                                float(i)
                                for i in line.lstrip().rstrip("\n").split(" ")
                                if i
                            ]
                        )
                    )
    spectral_bands = np.array(
        lines,
        dtype=[
            ("spectral_band_index", "u4"),
            ("lower_wavelength_limit", "f4"),
            ("upper_wavelength_limit", "f4"),
        ],
    )
    return spectral_bands



"""Runtime configuration."""

import threading

__all__ = ("RUNTIME",)


class RuntimeOpts(threading.local):
    """Run-time configuration controller."""

    def __init__(self, figsave_stamp=False, figsave_reduce_size=False):
        """Initialise run-time configuration."""
        self.__dict__["figsave_stamp"] = figsave_stamp
        self.__dict__["figsave_reduce_size"] = figsave_reduce_size

    def __repr__(self):
        """Repr of run-time configuration."""
        keyval_pairs = [f"{key}={self.__dict__[key]}" for key in self.__dict__]
        msg = f"RuntimeOpts({', '.join(keyval_pairs)})"
        return msg

    def __setattr__(self, name, value):
        """Set attributes."""
        if name not in self.__dict__:
            msg = f"'RuntimeOpts' object has no attribute {name!r}"
            raise AttributeError(msg)
        self.__dict__[name] = value


RUNTIME = RuntimeOpts()



"""Test synthobs submodule."""

from pathlib import Path

import iris
import iris.coords
import iris.cube
import numpy as np
import numpy.testing as npt
import pytest

from aeolus import synthobs

iris.FUTURE.datum_support = True
TST_DATA = Path(__file__).parent / "data" / "test_data"


@pytest.fixture(scope="module")
def example_trans_day():
    return iris.load_cube(TST_DATA / "netcdf" / "planet_transmission_day.nc")


@pytest.fixture(scope="module")
def example_trans_night():
    return iris.load_cube(TST_DATA / "netcdf" / "planet_transmission_night.nc")


def test_read_spectral_bands():
    expected_spectral_band_index = [1, 2, 3, 4, 5]
    expected_lower_wavelength_limit = [
        0.0010000000474974513,
        0.0005000000237487257,
        0.00033333332976326346,
        0.0002500000118743628,
        0.00019999999494757503,
    ]
    expected_upper_wavelength_limit = [
        0.009999999776482582,
        0.0010000000474974513,
        0.0005000000237487257,
        0.00033333332976326346,
        0.0002500000118743628,
    ]
    actual = synthobs.read_spectral_bands(
        TST_DATA / "spectral" / "sp_sw_500ir_bd_hatp11"
    )
    assert isinstance(actual, np.ndarray)
    assert actual.shape[0] == 500
    npt.assert_allclose(
        expected_spectral_band_index, actual["spectral_band_index"][:5]
    )
    npt.assert_allclose(
        expected_lower_wavelength_limit, actual["lower_wavelength_limit"][:5]
    )
    npt.assert_allclose(
        expected_upper_wavelength_limit, actual["upper_wavelength_limit"][:5]
    )


def test_read_normalized_stellar_flux():
    expected = [
        3.7919454902446414e-12,
        7.004032465118826e-09,
        1.939694804775627e-08,
        3.824304428690084e-08,
        6.35967865036946e-08,
    ]
    actual = synthobs.read_normalized_stellar_flux(
        TST_DATA / "spectral" / "sp_sw_500ir_bd_hatp11"
    )
    assert isinstance(actual, iris.cube.Cube)
    assert actual.shape[0] == 500
    npt.assert_allclose(expected, actual.data[:5])


def test_calc_stellar_flux():
    expected = [
        4.66409295e-10,
        8.61495993e-07,
        2.38582461e-06,
        4.70389445e-06,
        7.82240474e-06,
    ]
    actual = synthobs.calc_stellar_flux(
        TST_DATA / "spectral" / "sp_sw_500ir_bd_hatp11", 123
    )
    assert isinstance(actual, iris.cube.Cube)
    assert actual.shape[0] == 500
    npt.assert_allclose(expected, actual.data[:5])


def test_calc_geom_mean_mirrored(example_trans_day, example_trans_night):
    # Default
    actual = synthobs.calc_geom_mean_mirrored(
        example_trans_day, example_trans_night
    )
    npt.assert_allclose(actual.data.max(), 1.5901232591606386e-08)
    npt.assert_allclose(
        actual.data[123, 45, 102:108],
        [
            0.00000000e00,
            5.28650299e-40,
            1.09253009e-10,
            7.16136321e-10,
            5.28943014e-10,
            0.00000000e00,
        ],
    )
    assert actual.units == example_trans_day.units
    assert actual.shape == example_trans_day.shape
    # With an additional shift along the x-coordinate
    actual = synthobs.calc_geom_mean_mirrored(
        example_trans_day, example_trans_night, add_shift=-1
    )
    npt.assert_allclose(actual.data.max(), 1.76899099563551e-08)
    npt.assert_allclose(
        actual.data[123, 45, 102:109],
        [
            0.00000000e00,
            1.54026439e-67,
            2.12892023e-11,
            6.61482043e-10,
            7.37627923e-10,
            3.77198057e-10,
            0.00000000e00,
        ],
    )
    assert actual.units == example_trans_day.units
    assert actual.shape == example_trans_day.shape


def test_calc_transmission_spectrum(example_trans_day):
    expected_spectral_bands = [
        0.0055,
        0.00075,
        0.00041666668,
        0.00029166666,
        0.000225,
    ]
    expected_lower_wavelength_limit = [
        0.0010000000474974513,
        0.0005000000237487257,
        0.00033333332976326346,
        0.0002500000118743628,
        0.00019999999494757503,
    ]
    expected_upper_wavelength_limit = [
        0.009999999776482582,
        0.0010000000474974513,
        0.0005000000237487257,
        0.00033333332976326346,
        0.0002500000118743628,
    ]
    expected_rp_eff_over_rs = [
        0.06043149,
        0.0605353,
        0.06054769,
        0.060621,
        0.06062967,
    ]
    actual_rp_eff_over_rs = synthobs.calc_transmission_spectrum(
        example_trans_day,
        TST_DATA / "spectral" / "sp_sw_500ir_bd_hatp11",
        300.85538168664425,
        475026500.0,
        28995379.0,
    )
    actual_spectral_bands = actual_rp_eff_over_rs.coord(
        "spectral_band_centres"
    )
    actual_spectral_band_ll = actual_rp_eff_over_rs.coord(
        "spectral_band_lower_limit"
    )
    actual_spectral_band_ul = actual_rp_eff_over_rs.coord(
        "spectral_band_upper_limit"
    )
    assert isinstance(actual_spectral_bands, iris.coords.AuxCoord)
    assert isinstance(actual_rp_eff_over_rs, iris.cube.Cube)
    assert actual_spectral_bands.shape[0] == 500
    assert actual_rp_eff_over_rs.shape[0] == 500
    npt.assert_allclose(
        expected_spectral_bands, actual_spectral_bands.points[:5]
    )
    npt.assert_allclose(
        expected_lower_wavelength_limit, actual_spectral_band_ll.points[:5]
    )
    npt.assert_allclose(
        expected_upper_wavelength_limit, actual_spectral_band_ul.points[:5]
    )
    npt.assert_allclose(
        expected_rp_eff_over_rs, actual_rp_eff_over_rs.data[:5]
    )


def test_calc_transmission_spectrum_day_night_average(
    example_trans_day, example_trans_night
):
    # Default
    expected_rp_eff_over_rs = [
        0.1289366838346542,
        0.1289351109234838,
        0.1289354291294527,
        0.1289365978678745,
        0.1289363841112755,
    ]
    actual_rp_eff_over_rs = (
        synthobs.calc_transmission_spectrum_day_night_average(
            example_trans_day,
            example_trans_night,
            spectral_file=TST_DATA / "spectral" / "sp_sw_500ir_bd_hatp11",
            stellar_constant_at_1_au=1272.86475403,
            stellar_radius=7.302834e08,
            planet_top_of_atmosphere=94193200,
        )
    )
    assert isinstance(actual_rp_eff_over_rs, iris.cube.Cube)
    assert actual_rp_eff_over_rs.shape[0] == 500
    npt.assert_allclose(
        expected_rp_eff_over_rs, actual_rp_eff_over_rs.data[-5:]
    )
    # With an additional shift
    expected_rp_eff_over_rs = [
        0.1289324245796913,
        0.1289306558725381,
        0.1289309758113692,
        0.1289323585757668,
        0.1289320864595353,
    ]
    actual_rp_eff_over_rs = (
        synthobs.calc_transmission_spectrum_day_night_average(
            example_trans_day,
            example_trans_night,
            add_shift=-1,
            spectral_file=TST_DATA / "spectral" / "sp_sw_500ir_bd_hatp11",
            stellar_constant_at_1_au=1272.86475403,
            stellar_radius=7.302834e08,
            planet_top_of_atmosphere=94193200,
        )
    )
    npt.assert_allclose(
        expected_rp_eff_over_rs, actual_rp_eff_over_rs.data[-5:]
    )



"""Statistical functions."""

from collections.abc import Iterable, Mapping

import iris.analysis
from iris.analysis.maths import apply_ufunc
from iris.coords import AuxCoord
from iris.cube import Cube
from iris.exceptions import CoordinateCollapseError as CoColErr
from iris.exceptions import CoordinateNotFoundError as CoNotFound
import iris.util
from iris.util import broadcast_to_shape, promote_aux_coord_to_dim_coord
import numpy as np

from ..coord import area_weights_cube, coord_to_cube, ensure_bounds
from ..exceptions import ArgumentError, _warn
from ..model import um
from ..subset import (
    extract_after_n_days,
    extract_between_days,
    extract_last_n_days,
)
from .calculus import integrate

__all__ = (
    "abs_coord_mean",
    "after_n_day_mean",
    "between_day_mean",
    "cumsum",
    "last_n_day_mean",
    "meridional_mean",
    "minmaxdiff",
    "normalize_cube",
    "region_mean_diff",
    "spatial",
    "spatial_mean",
    "spatial_quartiles",
    "time_mean",
    "vertical_mean",
    "zonal_mean",
)


def abs_coord_mean(cube, coord):
    """
    Calculate cube's average over absolute values of a coordinate.

    For example, applying this to a cube with N latitudes ranging
    from SP to NP returns a cube with (N+1)//2 latitudes.

    Parameters
    ----------
    cube: iris.cube.Cube
        Cube with the specified coordinate.
    coord: str or iris.coords.Coord
        Coordinate to apply abs().

    Returns
    -------
    iris.cube.Cube
        Cube with a reduced dimension.
    """
    sign_lat = apply_ufunc(
        np.sign, coord_to_cube(cube, coord, broadcast=False)
    )
    sign_cube = sign_lat * cube
    _coord = sign_cube.coord(coord)
    _coord_dim = sign_cube.coord_dims(_coord)
    abs_coord = AuxCoord.from_coord(_coord)
    abs_coord.rename(f"abs_{abs_coord.name()}")
    abs_coord.points = np.abs(_coord.points)
    abs_coord.bounds = np.abs(_coord.bounds)
    sign_cube.add_aux_coord(abs_coord, data_dims=_coord_dim)
    out = sign_cube.aggregated_by(abs_coord, iris.analysis.MEAN)
    out.remove_coord(_coord)
    out.rename(cube.name())
    out.units = cube.units
    promote_aux_coord_to_dim_coord(out, abs_coord.name())
    out.coord(abs_coord).rename(_coord.name())
    return out


def cumsum(cube, axis, axis_weights=False, model=um):
    """
    Cumulative sum of a cube.

    Parameters
    ----------
    cube: iris.cube.Cube
        Input cube.
    axis: str
        Coordinate axis of operation (t|z|y|x).
    axis_weights: bool, optional
        If True, multiply data by the coordinate spacings.
    model: aeolus.model.Model, optional
        Model class with a relevant coordinate name.

    Returns
    -------
    iris.cube.Cube
        Cube of cumulative sums with the same dimensions as the input cube.
    """
    try:
        c = cube.coord(getattr(model, axis)).copy()
    except (AttributeError, CoNotFound):
        c = cube.coord(axis=axis).copy()
    dim = cube.coord_dims(c)
    if axis_weights:
        if not c.has_bounds():
            c.guess_bounds()
        weights = broadcast_to_shape(
            c.bounds[:, 1] - c.bounds[:, 0], cube.shape, dim
        )
        data = cube.data * weights
        units = cube.units * c.units
    else:
        data = cube.data
        units = cube.units
    data = np.nancumsum(data, axis=dim[0])
    res = cube.copy(data=data)
    res.rename(f"cumulative_sum_of_{cube.name()}_along_{axis}")
    res.units = units
    return res


def last_n_day_mean(cube, days=365, model=um):
    """Average the cube over the last `n` days of its time dimension."""
    cube_sub = time_mean(
        extract_last_n_days(cube, days=days, model=model), model=model
    )
    return cube_sub


def after_n_day_mean(cube, days=365, model=um):
    """Average the cube over the last `n` days of its time dimension."""
    cube_sub = time_mean(
        extract_after_n_days(cube, days=days, model=model), model=model
    )
    return cube_sub


def between_day_mean(cube, day_start, day_end, model=um):
    """Average the cube over a subset of days."""
    cube_sub = time_mean(
        extract_between_days(
            cube, day_start=day_start, day_end=day_end, model=model
        ),
        model=model,
    )
    return cube_sub


def meridional_mean(cube, model=um):
    """
    Calculate cube's meridional average.

    Parameters
    ----------
    cube: iris.cube.Cube
        Cube with a latitude coordinate.
    model: aeolus.model.Model, optional
        Model class with a relevant coordinate name.

    Returns
    -------
    iris.cube.Cube
        Collapsed cube.
    """
    lat_name = model.y
    coslat = np.cos(np.deg2rad(cube.coord(lat_name).points))
    coslat2d = broadcast_to_shape(
        coslat, cube.shape, cube.coord_dims(lat_name)
    )
    cube_mean = (cube * coslat2d).collapsed(
        lat_name, iris.analysis.SUM
    ) / np.sum(coslat)
    return cube_mean


def minmaxdiff(cubelist, name):
    """
    Spatial maximum minus spatial minimum for a given cube.

    Parameters
    ----------
    cubelist: iris.cube.CubeList
        Input list of cubes.
    name: str
        Cube name.

    Returns
    -------
    iris.cube.Cube
        Difference between the extrema with collapsed spatial dimensions.
    """
    _min = spatial(cubelist.extract_cube(name), "min")
    _max = spatial(cubelist.extract_cube(name), "max")
    diff = _max - _min
    diff.rename(f"{name}_difference")
    return diff


def normalize_cube(cube):
    """
    Normalize cube data, i.e. make the values range from 0 to 1.

    .. math::
        z_i = (x_i - min(x)) / (max(x) - min(x))

    Parameters
    ----------
    cube: iris.cube.Cube
        The input cube.
    """
    cube_min = cube.collapsed(cube.dim_coords, iris.analysis.MIN)
    cube_max = cube.collapsed(cube.dim_coords, iris.analysis.MAX)
    norm = (cube - cube_min) / (cube_max - cube_min)
    norm.rename(f"normalized_{cube.name()}")
    return norm


def region_mean_diff(cubelist, name, region_a, region_b):
    """
    Difference between averages over two regions for a given cube.

    Parameters
    ----------
    cubelist: iris.cube.CubeList
        Input list of cubes.
    name: str
        Cube name.
    region_a: aeolus.region.Region
        First region.
    region_b: aeolus.region.Region
        Second region.

    Returns
    -------
    iris.cube.Cube
        Difference between the region averages with collapsed spatial dims.
    """
    mean_a = spatial_mean(
        cubelist.extract_cube(name).extract(region_a.constraint)
    )
    mean_b = spatial_mean(
        cubelist.extract_cube(name).extract(region_b.constraint)
    )
    diff = mean_a - mean_b
    diff.rename(f"{name}_mean_diff_{region_a}_{region_b}")
    return diff


def spatial(cube, aggr, model=um):
    """
    Calculate spatial statistic with geographic grid weights.

    Parameters
    ----------
    cube: iris.cube.Cube
        Cube with longitude and latitude coordinates.
    aggr: str
        Statistical aggregator (see iris.analysis for available aggregators).
    model: aeolus.model.Model, optional
        Model class with relevant coordinate names.

    Returns
    -------
    iris.cube.Cube
        Collapsed cube.

    Examples
    --------
    >>> spatial(my_data_cube, "mean")

    """
    ensure_bounds(cube, coords=("x", "y"), model=model)
    coords = (model.y, model.x)
    flag = all(cube.coord(c).has_bounds() for c in coords)
    aggregator = getattr(iris.analysis, aggr.upper())
    if flag and isinstance(aggregator, iris.analysis.WeightedAggregator):
        kw = {"weights": area_weights_cube(cube, normalize=True).data}
    else:
        kw = {}
    try:
        out = cube.collapsed(coords, aggregator, **kw)
    except CoColErr as e:
        _warn(f"Caught exception in spatial():\n{e}")
        # out = iris.util.squeeze(cube)
        out = cube
    return out


def spatial_mean(cube, model=um):
    """Shortcut for spatial(cube, "mean")."""
    return spatial(cube, "mean", model=model)


def spatial_quartiles(cube, model=um):
    """Calculate quartiles over horizontal coordinates."""
    _warn("No weights are applied!")
    q25 = cube.collapsed(
        (model.y, model.x), iris.analysis.PERCENTILE, percent=25
    )
    q75 = cube.collapsed(
        (model.y, model.x), iris.analysis.PERCENTILE, percent=75
    )
    return q25, q75


def time_mean(obj, squeeze=False, model=um):
    """Time average of a cube or a container with cubes."""
    if isinstance(obj, Cube):
        try:
            out = obj.collapsed(model.t, iris.analysis.MEAN)
        except CoColErr as e:
            _warn(f"Caught exception in time_mean():\n{e}")
            out = obj
        if squeeze:
            out = iris.util.squeeze(out)
    elif isinstance(obj, Mapping):
        out = {}
        for key, cube in obj.items():
            out[key] = time_mean(cube, model=model)
        out = obj.__class__(out)
    elif isinstance(obj, Iterable):
        out = []
        for cube in obj:
            out.append(time_mean(cube, model=model))
        out = obj.__class__(out)
    else:
        raise ArgumentError(f"Unrecognised type of obj: {type(obj)}")
    return out


def vertical_mean(cube, weight_by=None, model=um):
    """
    Vertical mean of a cube with optional weighting.

    Parameters
    ----------
    cube: iris.cube.Cube
        Cube to average.
    weight_by: str or iris.coords.Coord or iris.cube.Cube, optional
        Coordinate of the given cube or another cube used for weighting.
    model: aeolus.model.Model, optional
        Model class with a relevant coordinate name.

    Returns
    -------
    iris.cube.Cube
        Collapsed cube.
    """
    coord = model.z
    if len(cube.coord_dims(coord)) == 0:
        _warn(
            f"{repr(coord)} does not describe any dimension in {repr(cube)}."
        )
        return cube
    if weight_by is None:
        vmean = cube.collapsed(coord, iris.analysis.MEAN)
    else:
        if isinstance(weight_by, (str, iris.coords.Coord)):
            weights = broadcast_to_shape(
                cube.coord(weight_by).points.squeeze(),
                cube.shape,
                cube.coord_dims(weight_by),
            )
            vmean = cube.collapsed(coord, iris.analysis.MEAN, weights=weights)
        elif isinstance(weight_by, iris.cube.Cube):
            a_copy = cube.copy()
            b_copy = weight_by.copy()
            a_copy.coord(coord).bounds = None
            b_copy.coord(coord).bounds = None
            prod = b_copy * a_copy
            vmean = integrate(prod, coord) / integrate(weight_by, coord)
        else:
            raise ArgumentError(
                f"Unrecognised type of weight_by: {type(weight_by)}"
            )
    vmean.rename(f"vertical_mean_of_{cube.name()}")
    return vmean


def zonal_mean(cube, model=um):
    """
    Calculate cube's zonal average.

    Parameters
    ----------
    cube: iris.cube.Cube
        Cube with a latitude coordinate.
    model: aeolus.model.Model, optional
        Model class with a relevant coordinate name.

    Returns
    -------
    iris.cube.Cube
        Collapsed cube.
    """
    lon_name = model.x
    cube_mean = cube.collapsed(lon_name, iris.analysis.MEAN)
    return cube_mean



"""Test coord submodule."""

from pathlib import Path

import iris.coords
import iris.cube
import iris.exceptions
import numpy as np
import numpy.testing as npt
import pytest

from aeolus import coord

iris.FUTURE.datum_support = True
TST_DATA = Path(__file__).parent / "data" / "test_data"


@pytest.fixture(scope="module")
def example_cubelist_from_file():
    return iris.load(str(TST_DATA / "netcdf" / "held_suarez_1200d.nc"))


def test_cell_centres():
    arr = np.array([-1, 0, 1, 2])
    des = np.array([-0.5, 0.5, 1.5])
    act = coord._cell_centres(arr)
    npt.assert_allclose(act, des)
    des = np.array([-0.8, 0.2, 1.2])
    act = coord._cell_centres(arr, bound_position=0.2)
    npt.assert_allclose(act, des)


def test_cell_bounds():
    arr = np.array([26.0, 27.0, 28.0, 29.0, 30.0])
    des = np.array([25.5, 26.5, 27.5, 28.5, 29.5, 30.5])
    act = coord._cell_bounds(arr)
    npt.assert_allclose(act, des)
    des = np.array([25.0, 27.0, 28.0, 29.0, 30.0, 31.0])
    act = coord._cell_bounds(arr, bound_position=1)
    npt.assert_allclose(act, des)


def test__is_longitude_global():
    assert coord._is_longitude_global(np.arange(0, 360, 2.5))
    assert coord._is_longitude_global(np.linspace(-180, 180, 10))
    assert not coord._is_longitude_global(np.arange(0, 180, 2.5))


def test_coord_to_cube():
    xc = iris.coords.DimCoord([-1, 2, 3], units="m", standard_name="longitude")
    yc = iris.coords.DimCoord(
        [10, 30, 50, 70], units="m", standard_name="latitude"
    )
    zc = iris.coords.DimCoord(
        [1000, 500], units="hPa", standard_name="air_pressure"
    )
    arr = np.arange(24).reshape((2, 4, 3))
    cube = iris.cube.Cube(
        data=arr,
        dim_coords_and_dims=[i[::-1] for i in [*enumerate((zc, yc, xc))]],
        standard_name="x_wind",
        units="m/s",
    )
    cc_bc = coord.coord_to_cube(cube, "longitude")
    assert cc_bc.ndim == 3
    assert cc_bc.coord("latitude") == cube.coord("latitude")
    cc = coord.coord_to_cube(cube, "longitude", broadcast=False)
    assert cc.ndim == 1
    assert cc.shape == xc.shape
    assert cc.standard_name == "longitude"
    assert cc.units == "m"
    npt.assert_allclose(cc.data, xc.points)


def test_isel_cube(example_cubelist_from_file):
    cube = example_cubelist_from_file[0]
    result = coord.isel(cube, "latitude", 11)
    assert result.shape == (7, 72)
    assert result.coord("latitude").points[0] == -44.0
    kgo = [
        279.34421964,
        304.32601524,
        314.35886024,
        357.66854755,
        443.61117028,
        575.30497115,
        732.65202309,
    ]
    npt.assert_allclose(result.data[:, 0], kgo)
    result = coord.isel(cube, cube.coords()[-1], 0)
    assert result.shape == (45, 72)
    with pytest.raises(IndexError):
        coord.isel(cube, "latitude", 123, skip_not_found=True)
    with pytest.raises(iris.exceptions.CoordinateNotFoundError):
        coord.isel(cube, "blah", 11)
    result = coord.isel(cube, "blah", 11, skip_not_found=True)
    assert result is cube


def test_isel_cubelist(example_cubelist_from_file):
    cl = example_cubelist_from_file
    result = coord.isel(cl, "latitude", 11)
    assert all(i.shape == (7, 72) for i in result)
    assert result[-1].coord("latitude").points[0] == -44.0
    result = coord.isel(cl, "blah", 11)
    assert result == cl
    with pytest.raises(iris.exceptions.CoordinateNotFoundError):
        result = coord.isel(cl, "blah", 11, skip_not_found=False)
    cube1 = cl[0].copy()
    cube1.remove_coord("latitude")
    cube1.remove_coord("longitude")
    cube2 = cl[1]
    result = coord.isel([cube1, cube2], "latitude", 11)
    assert len(result) == 2



"""Test the meta submodule."""

import iris.cube
import pytest

from aeolus import meta
from aeolus.const import init_const
from aeolus.const.const import ConstContainer
from aeolus.exceptions import ArgumentError


def test_const_from_attrs_default():
    @meta.const_from_attrs()
    def foo(arg, const=None):
        """Test function."""
        return const

    constants = init_const()
    cube = iris.cube.Cube([], attributes={"planet_conf": constants})
    result = foo(cube)
    assert isinstance(result, ConstContainer)
    result = foo(iris.cube.CubeList([cube]))
    assert isinstance(result, ConstContainer)
    with pytest.raises(ArgumentError):
        foo(0)
    with pytest.raises(ArgumentError):
        foo(iris.cube.Cube([]))
    with pytest.raises(ArgumentError):
        foo(iris.cube.Cube([]), const=1234)
    with pytest.raises(ArgumentError):
        foo(iris.cube.Cube([], attributes={"foo": constants}))
    with pytest.raises(ArgumentError):
        foo(iris.cube.Cube([], attributes={"planet_conf": 1234}))


def test_const_from_attrs_relax():
    @meta.const_from_attrs(strict=False)
    def bar(arg, const=None):
        """Test function."""
        return const

    constants = init_const()
    cube = iris.cube.Cube([], attributes={"planet_conf": constants})
    result = bar(cube)
    assert isinstance(result, ConstContainer)
    result = bar(iris.cube.CubeList([cube]))
    assert isinstance(result, ConstContainer)
    # with pytest.warns(AeolusWarning):
    #     bar(iris.cube.Cube([]))
    # with pytest.warns(AeolusWarning):
    #     bar(iris.cube.Cube([]), const=1234)
    # with pytest.warns(AeolusWarning):
    #     bar(iris.cube.Cube([], attributes={"foo": constants}))
    # with pytest.warns(AeolusWarning):
    #     bar(iris.cube.Cube([], attributes={"planet_conf": 1234}))


def test_copy_doc():
    def foo():
        """Make docstring."""
        pass

    @meta.copy_doc(foo)
    def bar():
        pass

    assert foo.__doc__ == bar.__doc__



"""Text-related formatting functions."""

import itertools
import math
import os
import string

from ..calc import spatial_mean
from ..exceptions import ArgumentError

__all__ = (
    "all_sim_file_label",
    "cube_minmeanmax_str",
    "fmt_lonlat",
    "subplot_label_generator",
    "tex2cf_units",
    "unit_format",
)


def all_sim_file_label(sims):
    """Make a shorter label for a list of labels with a common prefix."""
    pref = os.path.commonprefix(sims)
    return f"{pref}_{'_'.join([i.removeprefix(pref) for i in sims])}"


def cube_minmeanmax_str(
    cube,
    sep=" | ",
    eq_sign="=",
    fmt="auto",
    weight=True,
    labels=None,
    **kw_unit_format,
):
    """Return min, mean and max of an iris cube as a string."""
    # Compute the stats
    if labels is None:
        labels = ["min", "mean", "max"]
    _min = float(cube.data.min())
    if weight:
        _mean = float(spatial_mean(cube).data)
    else:
        _mean = float(cube.data.mean())
    _max = float(cube.data.max())
    # Assemble a string
    txts = []
    for label, num in zip(labels, [_min, _mean, _max]):
        if fmt == "auto":
            if (math.log10(abs(_mean)) < 0) or (math.log10(abs(_mean)) > 5):
                _str = f"{label}{eq_sign}{num:.0e}"
            else:
                _str = f"{label}{eq_sign}{round(num):.0f}"
        elif fmt == "pretty":
            _str = f"{label}{eq_sign}{unit_format(num, **kw_unit_format)}"
        else:
            _str = f"{label}{eq_sign}{num:{fmt}}"
        txts.append(_str)
    return sep.join(txts)


def fmt_lonlat(value, lon_or_lat, degree=False):
    r"""
    Convert longitude or latitude value to string with a hemisphere identifier.

    Parameters
    ----------
    value: int
        Value of longitude or latitude.
        Note that this function is only for integer values.
    lon_or_lat: str
        Longitude or latitude
    degree: bool, optional
        If true, a TeX degree symbol is included

    Returns
    -------
    str

    Examples
    --------
    >>> fmt_lonlat(-25, "lon")
    '25W'
    >>> fmt_lonlat(89, "lat", degree=True)
    '89$^\\degree$N'
    >>> fmt_lonlat(0, "lon")
    '0'
    """
    from LatLon23 import Latitude, Longitude  # noqa

    if lon_or_lat.lower().startswith("lat"):
        res = Latitude(value)
    elif lon_or_lat.lower().startswith("lon"):
        res = Longitude(value)
    else:
        raise ArgumentError(
            "2nd arg or the function should start with `lon` or `lat`"
        )
    out = res.to_string("%d%H")
    if degree:
        out = out[:-1] + r"$^\degree$" + out[-1]
    if value == 0:
        out = out[:-1]
    return out


def subplot_label_generator():
    """Return generator of alphabetic labelling of subplots."""
    for i in itertools.count(1):
        for p in itertools.product(string.ascii_lowercase, repeat=i):
            yield "".join(p)


def tex2cf_units(unit_str):
    """Convert a TeX string to a string that can be used in cf_units."""
    return (
        unit_str.replace("$", "")
        .replace("{", "")
        .replace("}", "")
        .replace("^", "**")
    )


def unit_format(
    value, unit="1", decimal_digits=1, precision=None, exponent=None
):
    r"""
    Return a string representation of a given number with units.

    Format the scientific notation of the given number for use with LaTeX,
    with specified number of significant decimal digits and precision
    (number of decimal digits to show).
    The exponent to be used can also be specified explicitly.

    Parameters
    ----------
    value: int or float
        Number to be formatted.
    unit: str or int, optional
        Units of the number, if any.
    decimal_digits: int, optional
        Significant decimal digits.
    precision: int, optional
        Number of decimal digits to show.
    exponent: int, optional
        Exponent of the resulting number.

    Returns
    -------
    str

    Examples
    --------
    >>> unit_format(-1.234e-5)
    '$-1.2\\times10^{-5}$'
    >>> unit_format(6)
    '$6.0$'
    >>> unit_format(1.234e5, unit="m s^{-1}")
    '$1.2\\times10^{5}$ $m$ $s^{-1}$'
    >>> unit_format(1.234, decimal_digits=1, precision=3)
    '$1.200$'
    >>> unit_format(1.234, decimal_digits=3, precision=1)
    '$1.2$'
    """
    if unit in ["", "1", 1]:
        unit_str = ""
    else:
        unit_str = rf'${str(unit).replace(" ", "$ $")}$'
        if "%" in unit_str:
            unit_str = unit_str.replace("%", r"\%")
    if value == 1:
        string = unit_str
    else:
        if precision is None:
            precision = decimal_digits
        if exponent is None:
            try:
                exponent = int(math.floor(math.log10(abs(value))))
            except ValueError:
                exponent = 0

        coeff = round(value / 10**exponent, decimal_digits)

        if exponent in [0, 1]:
            string = rf"${round(value, decimal_digits):.{precision}f}$"
        else:
            string = rf"${coeff:.{precision}f}\times10^{{{exponent:d}}}$"
        if not unit == "1":
            string += rf" {unit_str}"
    return string



"""Utilities to plot cloud-related output of UM."""

import iris
import matplotlib as mpl
import numpy as np

from ..exceptions import _warn
from .cm_custom import cloudtypes_cmap

__all__ = ("CloudPlotter", "cloudtypes_cmap")


class CloudPlotter:
    """Factory to create a plot of low, medium, high cloud area fractions."""

    _stash_items = {"lo": "m01s09i203", "me": "m01s09i204", "hi": "m01s09i205"}
    _scales = {"lo": 100, "me": 200, "hi": 400}
    cloud_labels = [
        "No\nCloud",
        "L",
        "M",
        "L + M",
        "H",
        "H + L",
        "H + M",
        "All",
    ]

    def __init__(self, cubelist):
        """Initialise CloudPlotter."""
        self.factor = 10.0  # scaling factor
        self.cubes = {}
        for key, stash in self._stash_items.items():
            try:
                self.cubes[key] = cubelist.extract_cube(
                    iris.AttributeConstraint(STASH=stash)
                )
            except iris.exceptions.ConstraintMismatchError:
                _warn(
                    f"Warning!\n{key} ({stash}) is not found in\n\n{cubelist}"
                )

    def scale_data(self, threshold=0.1):
        """
        Scale cloud levels & add them together to make a composite cloud field.

        Values below `threshold` are set to zero.
        """
        for i, (key, cube) in enumerate(self.cubes.items()):
            cp_cube = cube.copy(
                data=np.where(
                    cube.data >= threshold,
                    cube.data * self.factor + self._scales[key],
                    0.0,
                )
            )
            if i == 0:
                self.aggr_cld = cp_cube
            else:
                self.aggr_cld += cp_cube
        self.aggr_cld.rename("aggregated_cloud_fraction")
        self.aggr_cld.attributes["source"] = " + ".join(
            [
                (
                    f"({self._stash_items[k]}*{self.factor}"
                    f"+{self._scales[k]})"
                )
                for k in self.cubes.keys()
            ]
        )

    def _make_norm(self, cmap, nsteps):
        """
        Create `BoundaryNorm` for cloud levels.

        Uses `nsteps` gradations of colour and `cmap` colormap.
        Also returns `midpoints` used for colorbar labelling.
        """
        base = np.linspace(0, self.factor, nsteps) + 1
        assert nsteps > 0, "number of intervals should be positive"
        lvls = [base]  # no clouds
        lvls.append(base + self._scales["lo"])  # low clouds (101-110)
        lvls.append(base + self._scales["me"])  # medium clouds (201-210)
        lvls.append(
            base * 2 + self._scales["lo"] + self._scales["me"]
        )  # low + medium (302-320)
        lvls.append(base + self._scales["hi"])  # high clouds (401-410)
        lvls.append(
            base * 2 + self._scales["lo"] + self._scales["hi"]
        )  # high + low (502-520)
        lvls.append(
            base * 2 + self._scales["hi"] + self._scales["me"]
        )  # high + medium (602-620)
        lvls.append(
            base * 3
            + self._scales["lo"]
            + self._scales["me"]
            + self._scales["hi"]
        )  # high + medium + low (703-730)
        lvls.append([1000])
        midpoints = []
        if nsteps == 1:
            for lvl1, lvl2 in zip(lvls[:-1], lvls[1:]):
                midpoints.append(0.5 * (lvl1[-1] + lvl2[0]))
        else:
            for lvl in lvls[:-1]:
                if len(lvl) % 2 == 0:
                    midpoints.append(lvl[len(lvl) // 2])
                else:
                    midpoints.append(
                        0.5 * (lvl[len(lvl) // 2] + lvl[(len(lvl) // 2) + 1])
                    )
        midpoints = np.asarray(midpoints)
        norm = mpl.colors.BoundaryNorm(np.concatenate(lvls), cmap.N)
        return norm, midpoints

    def get_all(self, nsteps=10, cmap=None):
        """Get cloud types cube and kwargs for external plotting."""
        if cmap is None:
            cmap = cloudtypes_cmap  # colormap from a separate file
        norm, cb_ticks = self._make_norm(cmap=cmap, nsteps=nsteps)
        plt_kw = {"norm": norm, "cmap": cmap}
        cb_labels = self.cloud_labels
        return self.aggr_cld, plt_kw, cb_ticks, cb_labels

    def pcolormesh(
        self, ax, xy=(), nsteps=10, cmap=None, cb_kwargs=None, **pc_kwargs
    ):
        """
        Display composite cloud plot using `matplotlib.pyplot.pcolormesh`.

        Parameters
        ----------
        ax: matplotlib axes object
            Axes where to create the plot in
        xy: tuple, optional
            Sequence of X and Y coordinate arrays (passed to pcolormesh)
        nsteps: integer, optional
            Number of colour steps for each of cloud category
        cmap: matplotlib.colors.ListedColormap, optional
            Colormap instance. If not given, custom cloud colormap is used.
        cb_kwags: dict, optional
            Dictionary of kwargs passed to colorbar
        **pc_kwargs: optional
            additional pcolormesh arguments

        Returns
        -------
        h: `matplotlib.collections.QuadMesh`
            pcolormesh instance
        cb: `matplotlib.colorbar.Colorbar`
            colorbar
        """
        if cb_kwargs is None:
            cb_kwargs = {}
        fig = ax.figure
        if cmap is None:
            cmap = cloudtypes_cmap  # colormap from a separate file
        norm, midpoints = self._make_norm(cmap=cmap, nsteps=nsteps)
        h = ax.pcolormesh(
            *xy, self.aggr_cld.data, cmap=cmap, norm=norm, **pc_kwargs
        )
        if "cax" in cb_kwargs:
            ax_kw = {}
        else:
            ax_kw = {"ax": ax}
        cb = fig.colorbar(h, **ax_kw, **cb_kwargs)
        cb.set_ticks(midpoints)
        cb.set_ticklabels(self.cloud_labels)
        return h, cb



"""Operations in tidally-locked coordinates."""

import iris.analysis
from iris.analysis.cartography import _meshgrid, rotate_pole, rotate_winds
from iris.coord_systems import RotatedGeogCS
from iris.coords import AuxCoord
from iris.cube import Cube
from iris.util import promote_aux_coord_to_dim_coord
import numpy as np

from ..coord import get_xy_coords
from ..exceptions import BadCoordinateError
from ..model import um

__all__ = (
    "regrid_to_rotated_pole_coordinates",
    "regrid_to_tidally_locked_coordinates",
    "rotate_winds_to_tidally_locked_coordinates",
)


def regrid_to_rotated_pole_coordinates(cube, pole_lon, pole_lat, model=um):
    """
    Regrid a cube to rotated-pole coordinates.

    Note: non-dimensional coordinates are lost.

    Parameters
    ----------
    cube: iris.cube.Cube
        Input cube.
    pole_lon: float
        New North Pole longitude.
    pole_lat: float
        New North Pole latitude.
    model: aeolus.model.Model, optional
        Model class with relevant coordinate and variable names.
    """
    _cs = cube.coord_system()
    xcoord, ycoord = get_xy_coords(cube, model=model)
    lon2d, lat2d = _meshgrid(xcoord.points, ycoord.points)
    r_lons, r_lats = rotate_pole(lon2d, lat2d, pole_lon, pole_lat)

    r_lon_coord = AuxCoord(
        r_lons.flatten(),
        units=xcoord.units,
        standard_name=xcoord.standard_name,
        var_name=xcoord.var_name,
        coord_system=_cs,
    )
    r_lat_coord = AuxCoord(
        r_lats.flatten(),
        units=ycoord.units,
        standard_name=ycoord.standard_name,
        var_name=ycoord.var_name,
        coord_system=_cs,
    )
    non_xy_coords = [
        (i, cube.coord_dims(i))
        for i in cube.dim_coords
        if i not in [xcoord, ycoord]
    ]
    # assume latitude and longitude are the rightmost dimensions
    lat_dim = len(non_xy_coords)
    if lat_dim != cube.coord_dims(ycoord)[0]:
        raise BadCoordinateError(
            "Ensure latitude and longitude are the rightmost dimensions."
        )
    cube_flat = Cube(
        cube.core_data().reshape(
            (*cube.shape[:lat_dim], np.product(cube.shape[lat_dim:]))
        ),
        aux_coords_and_dims=[
            *non_xy_coords,
            (r_lat_coord, lat_dim),
            (r_lon_coord, lat_dim),
        ],
        units=cube.units,
    )
    for coord, _ in non_xy_coords:
        promote_aux_coord_to_dim_coord(cube_flat, coord)

    out = cube_flat.regrid(cube, iris.analysis.UnstructuredNearest())
    return out


def regrid_to_tidally_locked_coordinates(
    cube, pole_lon=0, pole_lat=0, model=um
):
    """
    Regrid a cube to tidally locked coordinates.

    The substellar and antistellar point become the north and south pole,
    respectively. By default, the substellar point is assumed to be at (0, 0).

    Parameters
    ----------
    cube: iris.cube.Cube
        Input cube.
    pole_lon: float, optional
        New North Pole longitude.
    pole_lat: float, optional
        New North Pole latitude.
    model: aeolus.model.Model, optional
        Model class with relevant coordinate and variable names.
    """
    return regrid_to_rotated_pole_coordinates(
        cube, pole_lon=pole_lon, pole_lat=pole_lat, model=model
    )


def rotate_winds_to_tidally_locked_coordinates(u, v, pole_lon=0, pole_lat=0):
    """
    Rotate the horizontal wind components to tidally locked coordinates.

    The substellar and antistellar point become the north and south pole,
    respectively. By default, the substellar point is assumed to be at (0, 0).

    Parameters
    ----------
    u: iris.cube.Cube
        :class:`iris.cube.Cube` that contains the x-component of the vector.
    v: iris.cube.Cube
        :class:`iris.cube.Cube` that contains the y-component of the vector.
    pole_lon: float, optional
        New North Pole longitude.
    pole_lat: float, optional
        New North Pole latitude.
    """
    tl_cs = RotatedGeogCS(pole_lon, pole_lat, ellipsoid=u.coord_system())
    tl_u, tl_v = rotate_winds(u, v, tl_cs)
    return tl_u, tl_v



"""Input and output functionality."""

from collections.abc import Generator, Sequence
from pathlib import Path
import re
from typing import Any, Optional, Union

import iris
from iris.coord_systems import GeogCS
from iris.coords import AuxCoord
from iris.cube import Cube, CubeList
from iris.fileformats.pp import EARTH_RADIUS
from iris.fileformats.um import structured_um_loading
import iris.pandas
import numpy as np
import pandas as pd

__all__ = (
    "create_dummy_cube",
    "get_filename_list",
    "load_conservation_diag",
    "load_data",
    "load_multidir",
    "load_vert_lev",
    "save_cubelist",
)

GLM_RUNID = r"umglaa"  # file prefix
GLM_FILE_REGEX = (
    GLM_RUNID + r".p[b,c,d,e]{1}[0]{6}(?P<timestamp>[0-9]{2,6})_00"
)


def create_dummy_cube(
    nlat: Optional[int] = None,
    nlon: Optional[int] = None,
    n_res: Optional[int] = None,
    endgame: Optional[bool] = True,
    grid_type: Optional[str] = "a",
    pm180: Optional[bool] = False,
) -> Cube:
    """
    Create a dummy 2D cube with given resolution compatible with the UM grid.

    Parameters
    ----------
    nlat: int, optional
        Number of points in latitude. Not needed if `n_res` is given.
    nlon: int, optional
        Number of points in longitude. Not needed if `n_res` is given.
    n_res: int, optional
        N-notation resolution.
        If given, `nlat` and `nlon` are calculated automatically.
    endgame: bool, optional
        Use ENDGame grid.
        If True, "A" grid starts at lat=-90+dlat/2, lon=dlon/2.
        If False, "A" grid starts at lat=-90, lon=0.
        ENDGame's "A" grid is NewDyn "B" grid and vice versa.
    grid_type: str, optional
        Type of the UM grid.
          - A is main UM grid
          - B is staggered lorenz wind grid
          - Cu is staggered U wind grid
          - Cv is staggered V wind grid
    pm180: bool, optional
        Use -/+180 instead of 0 to 360 for the longitude span.

    Returns
    -------
    cube: iris.cube.Cube
        A dummy 2D cube of zeros.

    Examples
    --------
    >>> create_dummy_cube(nlat=90, nlon=144)
    >>> create_dummy_cube(n=96)
    """
    assert grid_type.lower() in [
        "a",
        "b",
        "cu",
        "cv",
    ], f"Grid {grid_type} not valid."

    # Calculate the number of points given the N-notation resolution.
    if n_res is not None:
        nlon = 2 * n_res
        nlat = 3 * (n_res // 2)

    # Default grid set up is for ENDGame A Grid, equivalent to NewDyn B grid.
    if endgame:
        lat_shift = grid_type.lower() in ("b", "cv")
        lon_shift = grid_type.lower() in ("b", "cu")
    else:
        lat_shift = grid_type.lower() in ("a", "cu")
        lon_shift = grid_type.lower() in ("a", "cv")

    # Set coordinate system to match PP data.
    geog_cs = GeogCS(EARTH_RADIUS)

    # Latitude
    spacing = 180.0 / nlat
    if lat_shift:
        lower_bound = -90.0
        upper_bound = 90.0
        act_nlat = nlat + 1
    else:
        lower_bound = -90.0 + spacing / 2.0
        upper_bound = 90.0 - spacing / 2.0
        act_nlat = nlat
    lats = np.linspace(lower_bound, upper_bound, act_nlat)
    lat_coord = iris.coords.DimCoord(
        lats, standard_name="latitude", units="degrees", coord_system=geog_cs
    )
    lat_coord.guess_bounds()

    # Longitude
    spacing = 360.0 / nlon
    if lon_shift:
        lower_bound = 0.0
        upper_bound = 360.0 - spacing
    else:
        lower_bound = 0.0 + spacing / 2.0
        upper_bound = 360.0 - spacing / 2.0
    lons = np.linspace(lower_bound, upper_bound, nlon)
    if pm180:
        lons -= 180.0
    lon_coord = iris.coords.DimCoord(
        lons,
        standard_name="longitude",
        units="degrees",
        circular=True,
        coord_system=geog_cs,
    )
    lon_coord.guess_bounds()

    # Create a numpy array of zeros.
    data = np.zeros((len(lats), len(lons)), dtype=np.int32)

    # Assemble the cube.
    cube = Cube(
        data,
        dim_coords_and_dims=((lat_coord, 0), (lon_coord, 1)),
        var_name="dummy_cube",
        units="1",
    )

    return cube


def full_path_glob(path: Path) -> Generator:
    """Recursive glob, including directories with regex."""
    resolved = Path(path.absolute().parts[0])
    glob_parts = []
    to_path = True
    for part in path.parts[1:]:
        if to_path and ("*" in part):
            to_path = False
        if to_path:
            resolved /= part
        else:
            glob_parts.append(part)
    gen = resolved.rglob(path.anchor.join(glob_parts))
    return gen


def get_filename_list(
    path_to_dir: Path,
    glob_pattern: Optional[str] = f"{GLM_RUNID}*",
    ts_start: Optional[int] = 0,
    ts_end: Optional[int] = -1,
    every: Optional[int] = 1,
    regex: Optional[str] = GLM_FILE_REGEX,
    regex_key: Optional[str] = "timestamp",
    sort: Optional[bool] = True,
) -> Sequence[Path]:
    """Get a list of files with timestamps >= than start in a directory."""
    glob_gen = full_path_glob(path_to_dir / glob_pattern)
    fnames = []
    tstamps = {}
    for fpath in glob_gen:
        match = re.match(regex, fpath.name)
        if match:
            ts_num = int(match[regex_key])
            if (ts_num >= ts_start) and (ts_num % every == 0):
                if (ts_end == -1) or (ts_num <= ts_end):
                    fnames.append(fpath)
                    tstamps[fpath] = ts_num
    if sort:
        fnames = sorted(fnames, key=lambda x: tstamps[x])
    return fnames


def load_conservation_diag(
    fnames: Sequence[Union[Path, str]],
    convert_to_iris: Optional[bool] = True,
    drop_duplicates: Optional[bool] = True,
) -> Union[pd.DataFrame, CubeList]:
    """Load UM conservation diagnostics from a series of text files."""
    dset = pd.concat(
        pd.read_csv(fpath, header=None, sep=r"\s+") for fpath in fnames
    )
    if drop_duplicates:
        dset = dset.drop_duplicates()
    dset = (
        dset.rename(
            {
                0: "timestep",
                1: "total_atmosphere_mass",
                2: "total_axial_angular_momentum",
                3: "total_kinetic_energy",
            },
            axis="columns",
        )
        .sort_values(by="timestep")
        .set_index("timestep")
    )
    if convert_to_iris:
        dset = iris.pandas.as_cubes(dset)
        cube = dset.extract_cube("total_atmosphere_mass")
        cube.units = "kg"
        cube = dset.extract_cube("total_axial_angular_momentum")
        cube.units = "kg m**2 s**-1"
        cube = dset.extract_cube("total_kinetic_energy")
        cube.units = "kg m**2 s**-2"
    return dset


def load_data(files: Sequence, structured: Optional[bool] = False) -> CubeList:
    """Use `iris.load` with optional structured loading for PP files."""
    if structured:
        with structured_um_loading():
            cubes = iris.load(files)
    else:
        with iris.FUTURE.context(datum_support=True):
            cubes = iris.load(files)
    return cubes


def load_multidir(
    path_mask: str, labels: Sequence[str], label_name: Optional[str] = "run"
) -> CubeList:
    """Load cubelists from multiple directories and merge."""
    joint_cl = CubeList()
    for label in labels:
        cl = iris.load(str(path_mask).format(label))
        for cube in cl:
            cube.attributes["um_version"] = ""  # FIXME
            cube.add_aux_coord(AuxCoord([label], long_name=label_name))
            joint_cl.append(cube)
    return joint_cl.merge()


def load_vert_lev(
    path_to_file: Path, lev_type: Optional[str] = "theta"
) -> np.array:
    """
    Read data from the UM vertical levels file.

    Parameters
    ----------
    path_to_file: pathlib.Path
        Full path to the vertical levels file.
    lev_type: str, optional
        What levels to return: "theta" or "rho".

    Returns
    -------
    levs: numpy.array
        Array of height levels.
    """
    import f90nml  # noqa

    with path_to_file.open("r") as nml_file:
        nml = f90nml.read(nml_file)
        levs = (
            np.array(nml["vertlevs"][f"eta_{lev_type}"])
            * nml["vertlevs"]["z_top_of_model"]
        )
    return levs


def save_cubelist(
    cubelist: CubeList, path: Path, **aux_attrs: Optional[Any]
) -> None:
    """
    Save a cubelist w/o the `planet_conf` container to a file.

    Parameters
    ----------
    cubelist: iris.cube.CubeList
        Cube list to write to disk.
    path: str or pathlib.Path
        File path.
    aux_attrs: dict, optional
        Dictionary of additional attributes to save with the cubes.
    """
    # Remove planet_conf attribute before saving
    out = CubeList()
    old_attrs = []
    for cube in cubelist:
        old_attrs.append(cube.attributes.copy())
        new_attrs = {**cube.attributes, **aux_attrs}
        try:
            new_attrs.pop("planet_conf")
        except KeyError:
            pass
        cube.attributes = new_attrs
        out.append(cube)
    with iris.FUTURE.context(save_split_attrs=True):
        iris.save(out, str(path))
    # Restore original attributes
    for cube, attrs in zip(cubelist, old_attrs):
        cube.attributes = attrs



"""LFRic-specific dictionaries of variable names and coordinates."""

from .base import Model

__all__ = ("lfric",)


lfric = Model(
    # Coordinates
    t="time",
    z="level_height",
    s="atmosphere_hybrid_sigma_pressure_coordinate",
    lev="full_levels",
    y="latitude",
    x="longitude",
    # Variables
    u="u_in_w3",
    v="v_in_w3",
    w="w_in_wth",
    pres="pressure_in_wth",
    thta="theta",
    # exner="exner_pressure",
    exner="exner_in_wth",
    temp="temperature",
    t_sfc="grid_surface_temperature",
    dens="density",
    p_sfc="pmsl",
    ghgt="geopotential_height",
    toa_isr="sw_direct_toa",
    toa_osr="sw_up_toa",
    toa_olr="lw_up_toa",
    toa_osr_cs="sw_up_clear_toa_rts",
    toa_olr_cs="lw_up_clear_toa_rts",
    caf="cloud_amount_maxrnd",
    sfc_dn_sw="sw_down_surf",
    sfc_dn_lw="lw_down_surf",
    sfc_up_lw="lw_up_surf",
    sfc_up_sw="sw_up_surf",
    ls_rain="ls_prec",  # TODO: correct
    cv_rain="conv_prec",  # TODO: correct
)



"""Matplotlib-related utilities."""

from collections.abc import Sequence
from pathlib import Path
from typing import Optional, Union

from iris.cube import Cube
from matplotlib.axes._axes import Axes
from matplotlib.figure import Figure
from matplotlib.lines import Line2D
import matplotlib.pyplot as plt
import numpy as np
from PIL import Image

from ..coord import get_cube_rel_days
from ..model import lfric, um
from ..model.base import Model
from ..runtime import RUNTIME

__all__ = (
    "add_custom_legend",
    "capitalise",
    "figsave",
    "hcross",
    "linspace_pm1",
    "make_list_2d",
    "map_scatter",
    "timeseries_1d",
    "timeseries_2d",
)


def add_custom_legend(
    ax_or_fig: Union[Axes, Figure], styles_and_labels: dict, **leg_kw
) -> None:
    """
    Add a custom legend to a matplotlib axis or figure.

    Parameters
    ----------
    ax_or_fig: matplotlib.axes._subplots.AxesSubplot / matplotlib.figure.Figure
        Matplotlib object where to put the legend.
    styles_and_labels: dict
        Dictionary with labels as keys and a dictionary of plot
        keywords as values.
    leg_kw: dict, optional
        Keyword arguments passed to `legend()` function.

    Example
    -------
    >>> import matplotlib.pyplot as plt
    >>> ax = plt.axes()
    >>> my_dict = dict(foo=dict(color='C0', marker="X"),
                       bar=dict(color='C1', marker="o"))
    >>> add_custom_legend(ax, my_dict, loc=2, title="blah")

    """
    lines = [Line2D([0], [0], **style) for style in styles_and_labels.values()]
    leg = ax_or_fig.legend(lines, styles_and_labels.keys(), **leg_kw)
    try:
        if ax_or_fig.legend_ is not None:
            ax_or_fig.add_artist(leg)
    except AttributeError:
        pass


def capitalise(
    s: str, sep_old: Optional[str] = "_", sep_new: Optional[str] = " "
) -> str:
    """Split the string and capitalise each word."""
    return sep_new.join([i.capitalize() for i in s.split(sep_old)])


def hcross(
    cube: Cube,
    ax: Optional[Axes] = None,
    model: Optional[Model] = um,
    **kw_plt,
) -> Optional[Axes]:
    """Plot a horizontal cross-section aka lat-lon map of a 2D cube."""
    newax = False
    if ax is None:
        ax = plt.axes()
        newax = True
    fig = ax.figure
    lons = cube.coord(um.x).points
    lats = cube.coord(um.y).points
    mappable = ax.pcolormesh(lons, lats, cube.data, **kw_plt)
    fig.colorbar(mappable, ax=ax)
    if newax:
        return ax


def linspace_pm1(n: int) -> np.typing.ArrayLike:
    """Return 2n evenly spaced numbers from -1 to 1, always skipping 0."""
    seq = np.linspace(0, 1, n + 1)
    return np.concatenate([-seq[1:][::-1], seq[1:]])


def make_list_2d(
    list_x: Sequence[str],
    list_y: Sequence[str],
    transpose: Optional[bool] = False,
) -> list:
    """Create a nested list out of 2 given lists."""
    if transpose:
        return [[f"{key_y}-{key_x}" for key_x in list_x] for key_y in list_y]
    else:
        return [[f"{key_x}-{key_y}" for key_x in list_x] for key_y in list_y]


def map_scatter(
    cube: Cube,
    ax: Optional[Axes] = None,
    model: Optional[Model] = lfric,
    **kw_plt,
) -> Optional[Axes]:
    """Plot a lat-lon scatter plot of a 2D cube."""
    newax = False
    if ax is None:
        ax = plt.axes()
        newax = True
    fig = ax.figure
    # This doesn't work because lons and lats in the mesh are of size N+2
    # while the data array is of size N
    # lons, lats = cube.mesh.node_coords
    # lons, lats = lons.points, lats.points
    lons = cube.coord(model.x).points
    lats = cube.coord(model.y).points
    mappable = ax.scatter(lons, lats, c=cube.data, **kw_plt)
    fig.colorbar(mappable, ax=ax)
    if newax:
        return ax


def timeseries_1d(
    cube: Cube,
    ax: Optional[Axes] = None,
    model: Optional[Model] = um,
    **kw_plt,
) -> Optional[Axes]:
    """Plot time series of a 1D cube."""
    newax = False
    if ax is None:
        ax = plt.axes()
        newax = True
    days = get_cube_rel_days(cube, model=um)
    ax.plot(days, cube.data, **kw_plt)
    if newax:
        return ax


def timeseries_2d(
    cube: Cube,
    ax: Optional[Axes] = None,
    model: Optional[Model] = um,
    **kw_plt,
) -> Optional[Axes]:
    """Plot time series of a 2D cube."""
    newax = False
    if ax is None:
        ax = plt.axes()
        newax = True
    fig = ax.figure
    days = get_cube_rel_days(cube, model=um)
    z = cube.coord(um.z).points
    mappable = ax.pcolormesh(days, z, cube.data.T, **kw_plt)
    fig.colorbar(mappable, ax=ax)
    if newax:
        return ax


def figsave(fig: Figure, filename: Path, **kw_savefig) -> None:
    """Save figure and print relative path to it."""
    if RUNTIME.figsave_stamp:
        fig.suptitle(
            filename.name,
            x=0.5,
            y=0.05,
            ha="center",
            fontsize="xx-small",
            color="tab:grey",
            alpha=0.5,
        )
    save_dir = filename.absolute().parent
    save_dir.mkdir(parents=True, exist_ok=True)
    fig.savefig(filename, **kw_savefig)
    fmt = plt.rcParams["savefig.format"]
    fname_orig = filename.with_suffix(f".{fmt}")
    if RUNTIME.figsave_reduce_size:
        # See MetPy Mondays #273
        orig_copy = fname_orig.with_stem(f"{fname_orig.stem}_original")
        fname_orig.replace(orig_copy)
        with Image.open(orig_copy) as im_orig:
            im = im_orig.convert("P", palette=Image.Palette.ADAPTIVE)
            im.save(fname_orig)
        # Delete the original
        orig_copy.unlink()
    pth = Path.cwd()
    rel_path = None
    pref = ""
    for par in pth.parents:
        pref += ".." + pth.anchor
        try:
            rel_path = f"{pref}{fname_orig.relative_to(par)}"
            break
        except ValueError:
            pass
    if rel_path is not None:
        print(f"Saved to {rel_path}")
        print(f"Size: {fname_orig.stat().st_size / 1024:.1f} KB")



"""Plotting functions used with cartopy."""

import cartopy.crs as ccrs
from cartopy.mpl.geoaxes import GeoAxes
from matplotlib.transforms import offset_copy
from mpl_toolkits.axes_grid1 import AxesGrid

from .text import fmt_lonlat

__all__ = ("GeoAxesGrid", "label_global_map_gridlines")


class GeoAxesGrid(AxesGrid):
    """
    Grid of cartopy axes.

    A subclass of :class:`mpl_toolkits.axes_grid1.AxesGrid` representing
    a grid of maps with the same projection :class:`~cartopy.crs.Projection`.

    - `axes_class` is defined automatically
    - The :class:`AxesGrid` built-in labelling is always switched off,
      and instead a standard procedure of creating
      grid lines and labels should be used.
    """

    def __init__(self, fig, rect, nrows_ncols, projection, **axesgrid_kw):
        """
        Initialise GeoAxesGrid.

        Build a :class:`GeoAxesGrid` instance with a grid nrows*ncols
        :class:`GeoAxes` with a projection :class:`~cartopy.crs.Projection`
        in :class:`~matplotlib.figure.Figure` *fig* with
        *rect=[left, bottom, width, height]* (in
        :class:`~matplotlib.figure.Figure` coordinates) or
        the subplot position code (e.g., "121").
        """
        axesgrid_kw["axes_class"] = (GeoAxes, {"map_projection": projection})
        axesgrid_kw["label_mode"] = ""  # note the empty label_mode
        super().__init__(fig, rect, nrows_ncols, **axesgrid_kw)


def label_global_map_gridlines(
    fig,
    ax,
    xticks=None,
    yticks=None,
    xoff=-10,
    yoff=-10,
    degree=False,
    **text_kw,
):
    """
    Label gridlines of a global cartopy map.

    Parameters
    ----------
    fig: matplotlib.figure.Figure
        Figure object.
    ax: cartopy.mpl.geoaxes.GeoAxesSubplot
        Cartopy axes.
    xticks: array-like, optional
        Sequence of longitude ticks.
    yticks: array-like, optional
        Sequence of latitude ticks.
    xoff: float, optional
        Longitude label offset from the axis (units are points).
        If negative (by default), the labels are drawn at the east boundary,
        otherwise at the west boundary.
    yoff: float, optional
        Latitude label offset from the axis (units are points).
        If negative (by default), the labels are drawn at the south boundary,
        otherwise at the north boundary.
    degree: bool, optional
        Add a degree symbol to tick labels.
    **text_kw: dict, optional
        Label text properties.
    """
    # Define what boundary to use
    if yticks is None:
        yticks = []
    if xticks is None:
        xticks = []
    extent = ax.get_extent(crs=ccrs.PlateCarree())
    if xoff <= 0:
        xpos = extent[0]  # labels at the east boundary
    else:
        xpos = extent[1]  # labels at the west boundary
    if yoff <= 0:
        ypos = extent[2]  # labels at the south boundary
    else:
        ypos = extent[3]  # labels at the north boundary
    geodetic_trans = ccrs.Geodetic()
    xlab_kw = ylab_kw = {**{"va": "center", "ha": "center"}, **text_kw}
    for xtick in xticks:
        s = fmt_lonlat(xtick, "lon", degree=degree)
        text_transform = offset_copy(
            geodetic_trans._as_mpl_transform(ax),
            fig=fig,
            units="points",
            x=0,
            y=yoff,
        )
        ax.text(xtick, ypos, s, transform=text_transform, **xlab_kw)

    for ytick in yticks:
        s = fmt_lonlat(ytick, "lat", degree=degree)
        text_transform = offset_copy(
            geodetic_trans._as_mpl_transform(ax),
            fig=fig,
            units="points",
            x=xoff,
            y=0,
        )
        ax.text(xpos, ytick, s, transform=text_transform, **ylab_kw)



"""Test calc submodule."""

from pathlib import Path

from cf_units import Unit
import iris
import iris.coord_systems
import iris.coords
import iris.cube
import iris.exceptions
import numpy as np
import numpy.testing as npt
import pytest

from aeolus import calc
from aeolus.exceptions import AeolusWarning

iris.FUTURE.datum_support = True
TST_DATA = Path(__file__).parent / "data" / "test_data"


@pytest.fixture(scope="module")
def example_cubelist_from_file():
    return iris.load(TST_DATA / "netcdf" / "held_suarez_1200d.nc")


def test_integrate():
    xc = iris.coords.DimCoord([-1, 2, 3], units="m", standard_name="longitude")
    yc = iris.coords.DimCoord(
        [10, 30, 50, 70], units="m", standard_name="latitude"
    )
    zc = iris.coords.DimCoord(
        [1000, 500], units="hPa", standard_name="air_pressure"
    )
    arr = np.arange(24).reshape((2, 4, 3))
    cube = iris.cube.Cube(
        data=arr,
        dim_coords_and_dims=[i[::-1] for i in [*enumerate((zc, yc, xc))]],
        standard_name="x_wind",
        units="m/s",
    )
    x_int = calc.integrate(cube, "longitude")
    npt.assert_allclose(
        x_int.data,
        np.array([[3.0, 15.0, 27.0, 39.0], [51.0, 63.0, 75.0, 87.0]]),
    )
    assert x_int.units == Unit("m2/s")
    assert x_int.name() == "integral_of_x_wind_wrt_longitude"
    t_arr = np.array(
        [
            [-300000.0, -350000.0, -400000.0],
            [-450000.0, -500000.0, -550000.0],
            [-600000.0, -650000.0, -700000.0],
            [-750000.0, -800000.0, -850000.0],
        ]
    )
    p_int = calc.integrate(cube, "air_pressure")
    p_int.convert_units("W m-2")
    npt.assert_allclose(p_int.data, t_arr)


def test_spatial(example_cubelist_from_file):
    cube = example_cubelist_from_file[0]
    result = calc.spatial(cube, "max")
    assert isinstance(result, iris.cube.Cube)
    assert result.shape == (7,)
    assert len(result.coords()) == 8
    assert result.coord("longitude") == iris.coords.DimCoord(
        points=np.array([180.0], dtype=np.float32),
        bounds=np.array([[0.0, 360.0]]),
        standard_name="longitude",
        var_name="longitude",
        units="degrees",
        coord_system=iris.coord_systems.GeogCS(6371229.0),
    )
    assert result.cell_methods[0].method == "maximum"
    assert result.cell_methods[0].coord_names == ("latitude", "longitude")
    kgo = [
        312.88888771,
        315.46518042,
        321.68071322,
        395.68476945,
        490.07506524,
        613.5893357,
        754.39737563,
    ]
    npt.assert_allclose(result.data, kgo)
    result = calc.spatial(cube, "MIN")
    kgo = [
        259.68906506,
        273.29139143,
        298.464596,
        320.08245475,
        423.69745232,
        544.69610463,
        698.78197378,
    ]
    npt.assert_allclose(result.data, kgo)


def test_spatial_mean(example_cubelist_from_file):
    cube = example_cubelist_from_file[0]
    result = calc.spatial_mean(cube)
    assert result == calc.spatial(cube, "mean")
    kgo = [
        286.0241389900014,
        304.5367306861022,
        313.7161581899977,
        350.3582590713208,
        446.03368741766377,
        570.7474771484258,
        728.6475794834439,
    ]
    npt.assert_allclose(result.data, kgo)


def test_spatial_mean_twice(example_cubelist_from_file):
    cube = example_cubelist_from_file[0]
    result = calc.spatial_mean(cube)
    with pytest.warns(AeolusWarning):
        other = calc.spatial_mean(result)
    assert result == other


def test_time_mean_single():
    xc = iris.coords.DimCoord([0, 1], units="m", standard_name="longitude")
    tc = iris.coords.DimCoord(
        np.arange(10),
        units="hours since 1970-01-01 00:00:00",
        standard_name="time",
    )
    tc.guess_bounds()
    arr = np.arange(20).reshape((10, 2))
    cube = iris.cube.Cube(
        data=arr,
        dim_coords_and_dims=[i[::-1] for i in [*enumerate((tc, xc))]],
        standard_name="x_wind",
        units="m/s",
    )
    result = calc.time_mean(cube)
    assert result.shape == (2,)
    assert result.units == "m/s"
    assert result.standard_name == "x_wind"
    kgo = [9.0, 10.0]
    npt.assert_allclose(result.data, kgo)
    result2 = calc.time_mean(cube, squeeze=True)
    assert result2 == result


def test_time_mean_collection():
    xc = iris.coords.DimCoord([0, 1], units="m", standard_name="longitude")
    tc = iris.coords.DimCoord(
        np.arange(10),
        units="hours since 1970-01-01 00:00:00",
        standard_name="time",
    )
    tc.guess_bounds()
    arr = np.arange(20).reshape((10, 2))
    cube1 = iris.cube.Cube(
        data=arr,
        dim_coords_and_dims=[i[::-1] for i in [*enumerate((tc, xc))]],
        standard_name="x_wind",
        units="m/s",
    )
    cube2 = iris.cube.Cube(
        data=arr * 2,
        dim_coords_and_dims=[i[::-1] for i in [*enumerate((tc, xc))]],
        standard_name="air_pressure",
        units="Pa",
    )
    cont_cl = iris.cube.CubeList([cube1, cube2])
    result = calc.time_mean(cont_cl)
    assert isinstance(result, iris.cube.CubeList)
    assert result[0] == calc.time_mean(cube1)
    assert result[1] == calc.time_mean(cube2)
    cont_tuple = (cube1, cube2)
    result = calc.time_mean(cont_tuple)
    assert isinstance(result, tuple)
    assert result[0] == calc.time_mean(cube1)
    assert result[1] == calc.time_mean(cube2)
    cont_dict = {"a": cube1, "b": cube2}
    result = calc.time_mean(cont_dict)
    assert isinstance(result, dict)
    assert result["a"] == calc.time_mean(cube1)
    assert result["b"] == calc.time_mean(cube2)



"""Test calc submodule."""

from pathlib import Path

import iris.cube
import numpy.testing as npt
import pytest

from aeolus import const
from aeolus.exceptions import ArgumentError, LoadError

TST_DATA = Path(__file__).parent / "data" / "test_data"
CONST_FILE = "dummy"


def test_init_const_general():
    cnsts = const.init_const()
    assert str(cnsts).startswith("GeneralConstants")
    assert isinstance(cnsts, const.const.ConstContainer)
    for key in cnsts.__dataclass_fields__.keys():
        attr = getattr(cnsts, key)
        assert isinstance(attr, iris.cube.Cube)
        assert attr.ndim == 0  # XXX could be relaxed
    key = "stefan_boltzmann"
    assert key in cnsts.__dataclass_fields__
    cube = getattr(cnsts, key)
    assert cube.units == "W m-2 K-4"
    npt.assert_allclose(cube.data, 5.670367e-08)


def test_loaderror():
    with pytest.raises(LoadError):
        const.init_const(CONST_FILE)
    with pytest.raises(LoadError):
        const.init_const(
            CONST_FILE, directory=TST_DATA / "nonexistent_directory"
        )


def test_argumenterror():
    with pytest.raises(ArgumentError):
        const.init_const(CONST_FILE, directory=str(TST_DATA / "json"))


def test_init_const_custom():
    cnsts = const.init_const(CONST_FILE, directory=TST_DATA / "json")
    assert str(cnsts).startswith("DummyConstants")
    assert isinstance(cnsts, const.const.ConstContainer)
    for key in cnsts.__dataclass_fields__.keys():
        attr = getattr(cnsts, key)
        assert isinstance(attr, iris.cube.Cube)
        assert attr.ndim == 0  # XXX could be relaxed
    key = "my_constant"
    assert key in cnsts.__dataclass_fields__
    cube = getattr(cnsts, key)
    assert cube.units == "m s-1"
    npt.assert_allclose(cube.data, 123)



"""Test io submodule."""

from pathlib import Path

import iris.coords
import iris.cube
import numpy as np
import numpy.testing as npt

from aeolus import io

TST_DATA = Path(__file__).parent / "data" / "test_data"


def test_create_dummy_cube():
    # using n_res
    cube_n72 = io.create_dummy_cube(n_res=72)
    assert cube_n72.shape == (108, 144)
    assert cube_n72.coord("longitude").units == "degrees"
    assert cube_n72.coord("longitude").circular
    npt.assert_allclose(
        cube_n72.coord("longitude").points[:3], [1.25, 3.75, 6.25]
    )
    npt.assert_allclose(
        cube_n72.coord("latitude").points[:3],
        [-89.16666667, -87.5, -85.83333333],
    )
    # using nlat and nlon
    cube = io.create_dummy_cube(nlat=108, nlon=144)
    assert cube_n72 == cube
    # using other grids
    assert io.create_dummy_cube(
        n_res=72, endgame=False, grid_type="b"
    ) == io.create_dummy_cube(n_res=72, endgame=True, grid_type="a")


def test_load_vert_lev_theta():
    levs = io.load_vert_lev(TST_DATA / "vert_lev" / "vertlevs_L38_29t_9s_40km")
    assert isinstance(levs, np.ndarray)
    assert levs.shape == (39,)
    npt.assert_allclose(levs[0], 0)
    npt.assert_allclose(levs[-1], 39254.833576)
    levs2 = io.load_vert_lev(
        TST_DATA / "vert_lev" / "vertlevs_L38_29t_9s_40km", lev_type="theta"
    )
    npt.assert_allclose(levs, levs2)


def test_load_vert_lev_rho():
    levs = io.load_vert_lev(
        TST_DATA / "vert_lev" / "vertlevs_L38_29t_9s_40km", lev_type="rho"
    )
    assert levs.shape == (38,)
    npt.assert_allclose(levs[0], 9.9982061118072)
    npt.assert_allclose(levs[-1], 36081.76331548462)


def test_save_cubelist():
    xc = iris.coords.DimCoord([-1, 2, 3], units="m", standard_name="longitude")
    yc = iris.coords.DimCoord(
        [10, 30, 50, 70], units="m", standard_name="latitude"
    )
    zc = iris.coords.DimCoord(
        [1000, 500], units="hPa", standard_name="air_pressure"
    )
    arr = np.arange(24).reshape((2, 4, 3))
    cube = iris.cube.Cube(
        data=arr,
        dim_coords_and_dims=[i[::-1] for i in [*enumerate((zc, yc, xc))]],
        standard_name="x_wind",
        units="m/s",
        attributes={"title": "Test data"},
    )
    cl = iris.cube.CubeList([cube])
    fname = TST_DATA / "test_save_cubelist.nc"
    io.save_cubelist(cl, fname, dummy_attr="foo bar")
    # Test that only the original attributes are retained
    assert cl[0].attributes == {"title": "Test data"}
    # Test that the file exists
    assert fname.is_file()
    # Load and compare
    result = iris.load_cube(str(fname))
    assert result.units == "m/s"
    assert "title" in result.attributes
    assert "dummy_attr" in result.attributes
    npt.assert_allclose(result.data, cube.data)



"""Test model submodule."""

from aeolus.model import um, um_stash
from aeolus.model.base import Model


def test_model_um():
    assert isinstance(um, Model)
    assert isinstance(um.u, str)
    assert um.u == "x_wind"


def test_model_um_stash():
    assert isinstance(um_stash, Model)
    assert isinstance(um_stash.u, str)
    assert um_stash.u == "m01s00i002"


def test_model_dummy():
    dummy = Model(u="abc")
    assert isinstance(dummy.u, str)
    assert dummy.u == "abc"
    assert dummy.v is None



"""Test the core submodule."""

# import contextlib
# import itertools
# import shutil
# import tempfile
# from datetime import datetime
from pathlib import Path

# import iris
#
# import numpy as np
# import numpy.testing as npt
import pytest

#
# from aeolus import core
from aeolus.exceptions import LoadError

TST_DATA = Path(__file__).parent / "data"

# _counter = itertools.count()


def test_foo():
    pass


def test_loaderror():
    with pytest.raises(LoadError):
        raise LoadError



"""Model-specific dictionaries of variable names and coordinates."""

from dataclasses import dataclass
import textwrap

__all__ = "Model"


@dataclass
class Model:
    """Base class for model-specific names."""

    def __hash__(self):  # noqa
        return hash(id(self))

    # Coordinates
    t: str = None  # time
    fcst_ref: str = None  # forecast reference
    fcst_prd: str = None  # forecast period
    z: str = None  # height
    lev: str = None  # level number
    s: str = None  # sigma
    d: str = None  # depth
    p: str = None  # pressure level
    y: str = None  # latitude
    x: str = None  # longitude
    # Variables
    # Main
    u: str = None
    v: str = None
    w: str = None
    pres: str = None
    thta: str = None
    exner: str = None
    sh: str = None
    t_sfc: str = None
    p_sfc: str = None
    # Radiation
    toa_isr: str = None
    toa_olr: str = None
    toa_olr_cs: str = None
    toa_osr: str = None
    toa_osr_cs: str = None
    sfc_dn_lw: str = None
    sfc_dn_lw_cs: str = None
    sfc_dn_sw: str = None
    sfc_dn_sw_cs: str = None
    sfc_up_lw: str = None
    sfc_up_lw_cs: str = None
    sfc_up_sw: str = None
    sfc_up_sw_cs: str = None
    sfc_net_down_lw: str = None
    sfc_net_down_sw: str = None
    lw_up: str = None
    lw_up_forcing: str = None
    lw_dn_forcing: str = None
    sw_up: str = None
    lw_dn: str = None
    sw_dn: str = None
    lw_up_cs: str = None
    sw_up_cs: str = None
    lw_dn_cs: str = None
    sw_dn_cs: str = None
    # BL
    sfc_shf: str = None
    sfc_lhf: str = None
    sfc_evap: str = None
    sfc_x_wind_stress: str = None
    sfc_y_wind_stress: str = None
    # Extra physics
    temp: str = None
    dens: str = None
    ghgt: str = None
    rh: str = None
    # Precip & Cloud
    cld_ice_mf: str = None
    cld_liq_mf: str = None
    rain_mf: str = None
    cca_anvil: str = None
    ccw_rad: str = None
    cld_top_hgt: str = None
    cld_ice_v: str = None
    cld_liq_v: str = None
    cld_v: str = None
    caf: str = None
    caf_h: str = None
    caf_m: str = None
    caf_l: str = None
    caf_vl: str = None
    ppn: str = None
    ls_rain: str = None
    ls_snow: str = None
    cv_rain: str = None
    cv_snow: str = None
    # Increments of temperature
    dt_sw: str = None
    dt_sw_cs: str = None
    dt_lw: str = None
    dt_lw_cs: str = None
    dt_lsppn: str = None
    dt_bl: str = None
    dt_cv: str = None
    dt_lscld: str = None
    dt_adv: str = None
    # Increments of specific humidity
    dq_sw: str = None
    dq_lw: str = None
    dq_lsppn: str = None
    dq_bl: str = None
    dq_cv: str = None
    dq_lscld: str = None
    dq_adv: str = None
    # Increments of eastward wind
    du_bl: str = None
    du_cv: str = None
    du_solve: str = None
    du_adv: str = None
    du_total: str = None
    # Soil
    soil_moist: str = None
    # Eliassen-Palm flux
    ep_flux_y: str = None
    ep_flux_z: str = None
    ep_flux_div: str = None
    # Lightning
    light_flash_rate: str = None
    light_number: str = None

    def __repr__(self):
        """Override the repr method of the dataclass."""
        size = len(
            [
                _
                for _, v in self.__dataclass_fields__.items()
                if v.name is not None
            ]
        )
        return f"{self.__class__.__name__} [{size} fields]"

    def __str__(self):
        """Override the str method of the dataclass."""
        fields = [
            v.name
            for _, v in self.__dataclass_fields__.items()
            if v.name is not None
        ]
        size = len(fields)
        s = (
            f"{self.__class__.__name__} with {size} fields:"
            f"\n{', '.join(fields)}"
        )
        return "\n".join(textwrap.wrap(s, 100))



"""Calculations."""

from .calculus import d_dx, d_dy, d_dz, deriv, div_h, integrate
from .diag import (
    air_density,
    air_potential_temperature,
    air_pressure,
    air_temperature,
    bond_albedo,
    calc_derived_cubes,
    dry_lapse_rate,
    flux,
    geopotential_height,
    ghe_norm,
    greenhouse_effect,
    heat_redist_eff,
    horiz_wind_cmpnts,
    meridional_mass_streamfunction,
    precip_sum,
    sfc_net_energy,
    sfc_water_balance,
    sigma_p,
    superrotation_index,
    toa_cloud_radiative_effect,
    toa_eff_temp,
    toa_net_energy,
    water_path,
    wind_rot_div,
    wind_speed,
    zonal_mass_streamfunction,
)
from .flux_h import (
    horizontal_fluxes_through_region_boundaries,
    net_horizontal_flux_to_region,
)
from .stats import (
    abs_coord_mean,
    after_n_day_mean,
    between_day_mean,
    cumsum,
    last_n_day_mean,
    meridional_mean,
    minmaxdiff,
    normalize_cube,
    region_mean_diff,
    spatial,
    spatial_mean,
    spatial_quartiles,
    time_mean,
    vertical_mean,
    zonal_mean,
)
from .tl import (
    regrid_to_rotated_pole_coordinates,
    regrid_to_tidally_locked_coordinates,
    rotate_winds_to_tidally_locked_coordinates,
)

__all__ = (
    "abs_coord_mean",
    "after_n_day_mean",
    "air_density",
    "air_potential_temperature",
    "air_pressure",
    "air_temperature",
    "between_day_mean",
    "bond_albedo",
    "calc_derived_cubes",
    "cumsum",
    "d_dx",
    "d_dy",
    "d_dz",
    "deriv",
    "div_h",
    "dry_lapse_rate",
    "flux",
    "geopotential_height",
    "ghe_norm",
    "greenhouse_effect",
    "heat_redist_eff",
    "horiz_wind_cmpnts",
    "horizontal_fluxes_through_region_boundaries",
    "integrate",
    "last_n_day_mean",
    "meridional_mass_streamfunction",
    "meridional_mean",
    "minmaxdiff",
    "net_horizontal_flux_to_region",
    "normalize_cube",
    "precip_sum",
    "region_mean_diff",
    "region_mean_diff",
    "regrid_to_rotated_pole_coordinates",
    "regrid_to_tidally_locked_coordinates",
    "rotate_winds_to_tidally_locked_coordinates",
    "sfc_net_energy",
    "sfc_water_balance",
    "sigma_p",
    "spatial",
    "spatial_mean",
    "spatial_quartiles",
    "superrotation_index",
    "time_mean",
    "toa_cloud_radiative_effect",
    "toa_eff_temp",
    "toa_net_energy",
    "vertical_mean",
    "water_path",
    "wind_rot_div",
    "wind_speed",
    "zonal_mass_streamfunction",
    "zonal_mean",
)



"""Subset cubes using iris constraints."""

from datetime import timedelta
from itertools import combinations

from iris import Constraint
from iris.cube import CubeList

from .coord import get_cube_datetimes
from .model import um

__all__ = (
    "CM_INST_CONSTR",
    "CM_MEAN_CONSTR",
    "DimConstr",
    "extract_after_n_days",
    "extract_between_days",
    "extract_last_month",
    "extract_last_n_days",
    "l_range_constr",
    "unique_cubes",
)


def _select_inst(cube):
    """Return True if cube has no cell methods."""
    return len(cube.cell_methods) == 0


def _select_mean(cube):
    """Return True if cube has a "mean" cell method."""
    try:
        return cube.cell_methods[0].method == "mean"
    except IndexError:
        return False


def _dim_constr(*coords, strict=True):
    """Make an `iris.Constraint` from given dimensional coordinates."""
    coord_set = set(coords)

    def __cube_func(cube):  # noqa
        cube_dimcoords = {dc.name() for dc in cube.dim_coords}
        if strict:
            # Cube has exactly the same dim coords
            return cube_dimcoords == coord_set
        else:
            # Cube has these dim coords and possibly other
            return coord_set.issubset(cube_dimcoords)

    return Constraint(cube_func=__cube_func)


def l_range_constr(h_min, h_max, units="km", coord=um.z):
    """Make a constraint on length range."""
    if units == "km":
        factor = 1e-3
    else:
        factor = 1
    return Constraint(
        **{coord: lambda x: h_min <= (x.point * factor) <= h_max}
    )


def extract_last_month(cube, model=um):
    """Extract time slices within the last months of a cube."""
    dt = get_cube_datetimes(cube)[-1]
    return cube.extract(
        Constraint(
            **{
                model.t: lambda x: (x.point.year == dt.year)
                and (x.point.month == dt.month)
            }
        )
    )


def extract_last_n_days(cube, days=365, model=um):
    """Extract time slices within the last `n` days of its time dimension."""
    dt = get_cube_datetimes(cube)[-1]
    ndays_before = dt - timedelta(days=days)
    cube_sub = cube.extract(
        Constraint(**{model.t: lambda t: t.point > ndays_before})
    )
    return cube_sub


def extract_between_days(cube, day_start, day_end, model=um):
    """Subset between `day_start` and `day_end` of its time coordinate."""
    dt_s = get_cube_datetimes(cube)[0]
    lbound = dt_s + timedelta(days=day_start)
    ubound = dt_s + timedelta(days=day_end)
    cube_sub = cube.extract(
        Constraint(**{model.t: lambda t: lbound <= t.point <= ubound})
    )
    return cube_sub


def extract_after_n_days(cube, days=365, model=um):
    """Extract time slices after the given number of `days`."""
    dt = get_cube_datetimes(cube)[0]
    ndays_after = dt + timedelta(days=days)
    cube_sub = cube.extract(
        Constraint(**{model.t: lambda t: t.point > ndays_after})
    )
    return cube_sub


CM_INST_CONSTR = Constraint(cube_func=_select_inst)
CM_MEAN_CONSTR = Constraint(cube_func=_select_mean)


class _ModeDimConstr:
    """Class for storing constraints in `DimConstr` grouped by their mode."""

    def __init__(self, **kwargs):
        """Assign all keyword arguments to attributes of this class."""
        self.__dict__.update(**kwargs)


class DimConstr:
    """
    Container for strict or relaxed dimensional constraints.

    Examples
    --------
    Extract cubes that have y and x dimensional coordinates (among others):

    >>> dc = DimConstr()
    >>> cubelist.extract(dc.relax.yx)

    Extract cubes that only have model levels, y and x dimensions:

    >>> dc = DimConstr()
    >>> cubelist.extract(dc.strict.myx)
    """

    def __init__(self, model=um):
        """
        Initialise DimConstr.

        Parameters
        ----------
        model: aeolus.model.Model, optional
            Model class with relevant coordinate names.
        """
        abbr_aliases = {
            "t": "t",
            "z": "z",
            "m": "lev",
            "p": "p",
            "y": "y",
            "x": "x",
        }
        for mode in ["strict", "relax"]:
            attrs = {}
            for key in ("z", "m", "p"):
                for n in range(1, 5):
                    for seq in combinations(["t", key, "y", "x"], n):
                        model_seq = [abbr_aliases[letter] for letter in seq]
                        attrs["".join(seq)] = _dim_constr(
                            *[getattr(model, i) for i in model_seq],
                            strict=(mode == "strict"),
                        )
            setattr(self, mode, _ModeDimConstr(**attrs))


def unique_cubes(cubelist):
    """Remove duplicate cubes from `iris.cube.CubeList`."""
    if len(cubelist) > 1:
        out = CubeList([cubelist[0]])
        for src_cube in cubelist[1:]:
            exists = False
            for dest_cube in out:
                if src_cube == dest_cube:
                    exists = True
            if not exists:
                out.append(src_cube)
        return out
    else:
        return cubelist


class CellMethodConstraint(Constraint):
    """Provides a simple Cube-CellMethod based :class:`Constraint`."""

    _names = ("method", "coord_names", "intervals", "comments")

    def __init__(
        self, method, coord_names=None, intervals=None, comments=None
    ):
        """
        Initialise a CellMethod-based constraint.

        Provides a Cube cell method based :class:`Constraint`, which matches
        against attributes of each of the cell methods in the cube, if there
        are any. See :class:`iris.coords.CellMethod` for more info.

        The constraint will succeed if *any* of the provided attributes match.

        Example usage::

            iris.CellMethodConstraint(method='mean')

            iris.CellMethodConstraint(method='mean',
                coord_names=("mesh_coordinates",))

        .. note:: Cell method names are case sensitive.

        """
        self.method = method
        self.coord_names = coord_names
        self.intervals = intervals
        self.comments = comments
        # self.cell_method = iris.coords.CellMethod(
        #     method, coords=coord_names,
        #     intervals=intervals, comments=comments
        # )
        super().__init__(cube_func=self._cube_func)

    def __eq__(self, other):
        eq = type(other) is CellMethodConstraint and all(
            getattr(self, attname) == getattr(other, attname)
            for attname in self._names
        )
        return eq

    def __hash__(self):
        # Must re-define if you overload __eq__ : Use object identity.
        return id(self)

    def _cube_func(self, cube):
        match = False
        try:
            cell_methods = cube.cell_methods
            for cell_method in cell_methods:
                for name in self._names:
                    expected = getattr(self, name)
                    actual = getattr(cell_method, name)
                    # Currently relaxed / permissive condition
                    if actual == expected:
                        match = True
                        break
        except AttributeError:
            match = False
        return match

    def __repr__(self):
        names = []
        for name in self._names:
            value = getattr(self, name)
            if value != "none":
                names.append(f"{name}={value!r}")
        return "{}({})".format(self.__class__.__name__, ", ".join(names))



"""Contains functionality for making matplotlib plots."""

from .cart import GeoAxesGrid, label_global_map_gridlines
from .cloud import CloudPlotter
from .cm_custom import cloudtypes_cmap
from .mpl import (
    add_custom_legend,
    capitalise,
    figsave,
    hcross,
    linspace_pm1,
    make_list_2d,
    map_scatter,
    timeseries_1d,
    timeseries_2d,
)
from .text import (
    all_sim_file_label,
    cube_minmeanmax_str,
    fmt_lonlat,
    subplot_label_generator,
    tex2cf_units,
    unit_format,
)

__all__ = (
    "add_custom_legend",
    "all_sim_file_label",
    "capitalise",
    "CloudPlotter",
    "cloudtypes_cmap",
    "cube_minmeanmax_str",
    "figsave",
    "fmt_lonlat",
    "GeoAxesGrid",
    "hcross",
    "label_global_map_gridlines",
    "linspace_pm1",
    "make_list_2d",
    "map_scatter",
    "subplot_label_generator",
    "tex2cf_units",
    "timeseries_1d",
    "timeseries_2d",
    "unit_format",
)



"""UM-specific dictionaries of variable names and coordinates."""

from .base import Model

__all__ = ("um", "um_stash")


um = Model(
    # Coordinates
    t="time",
    fcst_ref="forecast_reference_time",
    fcst_prd="forecast_period",
    z="level_height",
    lev="model_level_number",
    s="sigma",
    d="depth",
    p="air_pressure",
    y="latitude",
    x="longitude",
    # Variables
    u="x_wind",
    v="y_wind",
    w="upward_air_velocity",
    pres="air_pressure",
    thta="air_potential_temperature",
    exner="dimensionless_exner_function",
    sh="specific_humidity",
    t_sfc="surface_temperature",
    p_sfc="surface_air_pressure",
    toa_isr="toa_incoming_shortwave_flux",
    toa_olr="toa_outgoing_longwave_flux",
    toa_olr_cs="toa_outgoing_longwave_flux_assuming_clear_sky",
    toa_osr="toa_outgoing_shortwave_flux",
    toa_osr_cs="toa_outgoing_shortwave_flux_assuming_clear_sky",
    sfc_dn_lw="surface_downwelling_longwave_flux_in_air",
    sfc_dn_lw_cs="surface_downwelling_longwave_flux_in_air_assuming_clear_sky",
    sfc_dn_sw="surface_downwelling_shortwave_flux_in_air",
    sfc_dn_sw_cs="surface_downwelling_shortwave_flux_in_air_assuming_clear_sky",
    sfc_up_lw="surface_upwelling_longwave_flux_in_air",
    sfc_up_lw_cs="surface_upwelling_longwave_flux_in_air_assuming_clear_sky",
    sfc_up_sw="surface_upwelling_shortwave_flux_in_air",
    sfc_up_sw_cs="surface_upwelling_shortwave_flux_in_air_assuming_clear_sky",
    sfc_net_down_lw="surface_net_downward_longwave_flux",
    sfc_net_down_sw="surface_net_downward_shortwave_flux",
    lw_up="upwelling_longwave_flux_in_air",
    lw_up_forcing="upwelling_longwave_flux_in_air_with_forcing",
    lw_dn_forcing="downwelling_longwave_flux_in_air_with_forcing",
    sw_up="upwelling_shortwave_flux_in_air",
    lw_dn="downwelling_longwave_flux_in_air",
    sw_dn="downwelling_shortwave_flux_in_air",
    lw_up_cs="upwelling_longwave_flux_in_air_assuming_clear_sky",
    sw_up_cs="upwelling_shortwave_flux_in_air_assuming_clear_sky",
    lw_dn_cs="downwelling_longwave_flux_in_air_assuming_clear_sky",
    sw_dn_cs="downwelling_shortwave_flux_in_air_assuming_clear_sky",
    sfc_shf="surface_upward_sensible_heat_flux",
    sfc_lhf="surface_upward_latent_heat_flux",
    sfc_evap="surface_upward_water_flux",
    sfc_x_wind_stress="surface_downward_eastward_stress",
    sfc_y_wind_stress="surface_downward_northward_stress",
    temp="air_temperature",
    dens="air_density",
    ghgt="geopotential_height",
    rh="relative_humidity",
    cld_ice_mf="mass_fraction_of_cloud_ice_in_air",
    cld_liq_mf="mass_fraction_of_cloud_liquid_water_in_air",
    rain_mf="mass_fraction_of_rain_in_air",
    cca_anvil="convective_cloud_amount_with_anvil",
    ccw_rad="convective_cloud_water_passed_to_radiation",
    cld_top_hgt="total_cloud_top_height",
    cld_ice_v="ice_cloud_volume_fraction_in_atmosphere_layer",
    cld_liq_v="liquid_cloud_volume_fraction_in_atmosphere_layer",
    cld_v="cloud_volume_fraction_in_atmosphere_layer",
    caf="cloud_area_fraction_assuming_maximum_random_overlap",
    caf_h="high_type_cloud_area_fraction",
    caf_l="low_type_cloud_area_fraction",
    caf_m="medium_type_cloud_area_fraction",
    caf_vl="m01s09i202",
    ppn="precipitation_flux",
    ls_rain="stratiform_rainfall_flux",
    ls_snow="stratiform_snowfall_flux",
    cv_rain="convective_rainfall_flux",
    cv_snow="convective_snowfall_flux",
    dt_sw="change_over_time_in_air_temperature_due_to_shortwave_heating",
    dt_sw_cs="tendency_of_air_temperature_due_to_shortwave_heating_assuming_clear_sky",
    dt_lw="change_over_time_in_air_temperature_due_to_longwave_heating",
    dt_lw_cs="tendency_of_air_temperature_due_to_longwave_heating_assuming_clear_sky",
    dt_lsppn="change_over_time_in_air_temperature_due_to_stratiform_precipitation",
    dt_bl="change_over_time_in_air_temperature_due_to_boundary_layer_mixing",
    dt_cv="change_over_time_in_air_temperature_due_to_convection",
    dt_lscld="m01s09i181",
    dt_adv="change_over_time_in_air_temperature_due_to_advection",
    dq_sw="change_over_time_in_specific_humidity_due_to_shortwave_heating",
    dq_lw="change_over_time_in_specific_humidity_due_to_longwave_heating",
    dq_lsppn="change_over_time_in_specific_humidity_due_to_stratiform_precipitation",
    dq_bl="change_over_time_in_specific_humidity_due_to_boundary_layer_mixing",
    dq_cv="change_over_time_in_specific_humidity_due_to_convection",
    dq_lscld="m01s09i182",
    dq_adv="change_over_time_in_specific_humidity_due_to_advection",
    du_bl="change_over_time_in_x_wind_due_to_boundary_layer_mixing",
    du_cv="change_over_time_in_x_wind_due_to_convection",
    du_solve="change_over_time_in_x_wind_due_to_pressure_solver",
    du_adv="change_over_time_in_x_wind_due_to_advection",
    du_total="change_over_time_in_x_wind",
    soil_moist="moisture_content_of_soil_layer",
    # Eliassen-Palm flux
    ep_flux_y="northward_eliassen_palm_flux_in_air",
    ep_flux_z="upward_eliassen_palm_flux_in_air",
    ep_flux_div="tendency_of_eastward_wind_due_to_eliassen_palm_flux_divergence",
    # Lightning
    light_flash_rate="lightning_flash_rate",
    light_number="Number_of_lightning_flashes",
)

um_stash = Model(
    # Coordinates
    t="time",
    fcst_ref="forecast_reference_time",
    fcst_prd="forecast_period",
    z="level_height",
    lev="model_level_number",
    s="sigma",
    d="depth",
    p="air_pressure",
    y="latitude",
    x="longitude",
    # Variables
    u="m01s00i002",
    v="m01s00i003",
    w="m01s00i150",
    pres="m01s00i408",
    thta="m01s00i004",
    exner="m01s00i255",
    sh="m01s00i010",
    t_sfc="m01s00i024",
    p_sfc="m01s00i409",
    toa_isr="m01s01i207",
    toa_olr="m01s02i205",
    toa_olr_cs="m01s02i206",
    toa_osr="m01s01i208",
    toa_osr_cs="m01s01i209",
    sfc_dn_lw="m01s02i207",
    sfc_dn_lw_cs="m01s02i208",
    sfc_dn_sw="m01s01i235",
    sfc_dn_sw_cs="m01s01i210",
    sfc_net_down_lw="m01s02i201",
    sfc_net_down_sw="m01s01i201",
    lw_up="m01s02i217",
    lw_up_forcing="m01s02i417",
    lw_dn_forcing="m01s02i418",
    sw_up="m01s01i217",
    lw_dn="m01s02i218",
    sw_dn="m01s01i218",
    lw_up_cs="m01s02i219",
    sw_up_cs="m01s01i219",
    lw_dn_cs="m01s02i220",
    sw_dn_cs="m01s01i220",
    sfc_shf="m01s03i217",
    sfc_lhf="m01s03i234",
    temp="m01s03i236",
    rh="m01s30s113",
    cld_ice_mf="m01s00i012",
    cld_liq_mf="m01s00i254",
    rain_mf="m01s00i272",
    cca_anvil="m01s00i211",
    ccw_rad="m01s00i212",
    cld_top_hgt="m01s09i223",
    cld_ice_v="m01s00i268",
    cld_liq_v="m01s00i267",
    cld_v="m01s00i266",
    caf="m01s09i217",
    caf_h="m01s09i205",
    caf_m="m01s09i204",
    caf_l="m01s09i203",
    caf_vl="m01s09i202",
    ppn="m01s05i216",
    ls_rain="m01s04i203",
    ls_snow="m01s04i204",
    cv_rain="m01s05i205",
    cv_snow="m01s05i206",
    dt_sw="m01s01i181",
    dt_sw_cs="m01s01i233",
    dt_lw="m01s02i181",
    dt_lw_cs="m01s02i233",
    dt_lsppn="m01s04i181",
    dt_bl="m01s03i181",
    dt_cv="m01s05i181",
    dt_lscld="m01s09i181",
    dt_adv="m01s12i181",
    dq_sw="m01s01i182",
    dq_lw="m01s02i182",
    dq_lsppn="m01s04i182",
    dq_bl="m01s03i182",
    dq_cv="m01s05i182",
    dq_lscld="m01s09i182",
    dq_adv="m01s12i182",
    soil_moist="m01s08i223",
    ep_flux_y="m01s30i312",
    ep_flux_z="m01s30i313",
    ep_flux_div="m01s30i314",
    du_bl="m01s03i185",
    du_cv="m01s05i185",
    du_solve="m01s10i185",
    du_adv="m01s12i185",
    du_total="m01s30i185",
    light_flash_rate="m01s21i100",
    light_number="m01s21i104",
)



"""Logging utilities."""

from datetime import datetime
from pathlib import Path
import sys
from typing import Optional, Union

import loguru
from loguru import logger

__all__ = ("create_logger",)


def create_logger(
    script_path: Path, subdir: Optional[Union[str, Path]] = "logs"
) -> loguru._logger.Logger:
    """Create a logger using loguru."""
    logpath = script_path.parent / subdir
    logpath.mkdir(exist_ok=True)
    logger.configure(
        handlers=[
            {"sink": sys.stdout, "level": "INFO"},
            {
                "sink": logpath
                / (
                    f"log_{script_path.stem}"
                    f"_{datetime.now():%Y-%m-%d_%H%M%S}.log"
                ),
                "level": "DEBUG",
            },
        ]
    )
    return logger



"""Core submodule of aeolus package."""

from cached_property import cached_property
from iris.coord_systems import GeogCS
from iris.exceptions import ConstraintMismatchError as ConMisErr

from .calc import diag
from .const import add_planet_conf_to_cubes, init_const
from .coord import CoordContainer
from .decor import ReprAtmoSimBase
from .exceptions import _warn
from .meta import copy_doc
from .model import um
from .region import Region
from .subset import DimConstr

__all__ = ("AtmoSim", "AtmoSimBase")


class AtmoSimBase:
    """
    Base class for creating atmospheric model simulation classes in aeolus.

    Used to store and calculate atmospheric fields from gridded model output.
    Derived quantities are stored as cached properties to save time.

    Assumes the data are in spherical coordinates on a regular lat-lon grid.

    Attributes
    ----------
    name: str
        The run's name.
    description: str
        A description of the run.
    const: aeolus.const.ConstContainer
        Physical constants used in calculations for this run.
    model: aeolus.model.Model, optional
        Model class with relevant coordinate and variable names.
    """

    def __init__(
        self,
        cubes=None,
        name="",
        description="",
        planet="",
        const_dir=None,
        model=um,
        model_type=None,
        timestep=None,
        vert_coord=None,
        coord_check=False,
    ):
        """
        Instantiate an `AtmoSimBase` object.

        Parameters
        ----------
        cubes: iris.cube.CubeList
            Atmospheric fields.
        name: str, optional
            The name or label of this `AtmoSim`.
        description: str, optional
            This is not used internally; solely for the user's information.
        planet: str, optional
            Planet configuration used to get appropriate physical constants.
            If not given, Earth physical constants are initialised.
        const_dir: pathlib.Path, optional
            Path to a folder with files with constants for a specific planet.
        model: aeolus.model.Model, optional
            Model class with relevant coordinate and variable names.
        model_type: str, optional
            Type of the model run, global or LAM.
        timestep: int, optional
            Model time step in s.
        vert_coord: str, optional
            Character identificator for the model's vertical coordinate.
            "z" - data on "level_height"
            "p" - data on pressure levels
        coord_check: bool, optional
            Check if all cubes have the same set of coordinates.

        See Also
        --------
        aeolus.const.init_const, aeolus.coord.CoordContainer
        """
        self._cubes = cubes
        self.name = name
        self.description = description

        # Planetary constants
        self._update_planet(planet=planet, const_dir=const_dir)
        self._add_planet_conf_to_cubes()  # TODO: remove?

        # Model-specific names of variables and coordinates
        self.model = model
        self.dim_constr = DimConstr(model=self.model)

        # If the model is global or LAM (nested) and what its driving model is
        self.model_type = model_type
        self.timestep = timestep

        # Domain
        try:
            cube_yx = self._cubes.extract(self.dim_constr.relax.yx)[0]
            self.domain = Region.from_cube(
                cube_yx, name=f"{name}_domain", shift_lons=True
            )
        except IndexError:
            _warn("Initialised without a domain.")
            self.domain = None

        # Variables as attributes
        self._assign_fields()

        # Common coordinates
        dim_seq = ["tyx", "yx"]
        self.vert_coord = vert_coord
        # TODO: make it more flexible
        if self.vert_coord == "z":
            _constr = self.dim_constr.relax.z
        elif self.vert_coord == "p":
            _constr = self.dim_constr.relax.p
        else:
            _constr = self.dim_constr.relax.yx
        self.coord = CoordContainer(
            self._cubes.extract(_constr), coord_check=coord_check
        )

        if self.vert_coord is not None:
            dim_seq += [f"t{self.vert_coord}yx", f"{self.vert_coord}yx"]
        for seq in dim_seq:
            try:
                setattr(
                    self,
                    f"_ref_{seq}",
                    self._cubes.extract(getattr(self.dim_constr.strict, seq))[
                        0
                    ],
                )
            except IndexError:
                pass

        self.repr = ReprAtmoSimBase(self)

    def __repr__(self):  # noqa
        return self.repr.str_repr(short=True)

    def __str__(self):  # noqa
        return self.repr.str_repr(short=False)

    def _repr_html_(self):
        return self.repr.html_repr()

    @classmethod
    def from_parent_class(cls, obj):
        """Dynamically inherit from a similar class."""
        new_obj = cls(
            cubes=obj._cubes,
            name=obj.name,
            description=obj.description,
            planet=obj.planet,
            const_dir=obj.const_dir,
            model=obj.model,
            model_type=obj.model_type,
            timestep=obj.timestep,
            vert_coord=obj.vert_coord,
        )
        return new_obj

    def __getitem__(self, key):
        """Redirect self[key] to self.key."""
        return self.__getattribute__(key)

    def _assign_fields(self):
        """Assign input cubelist items as attributes of this class."""
        self.vars = []
        kwargs = {}
        for key in self.model.__dataclass_fields__:
            try:
                kwargs[key] = self._cubes.extract_cube(
                    getattr(self.model, key)
                )
                self.vars.append(key)
            except ConMisErr:
                pass
        self.__dict__.update(**kwargs)
        del kwargs, key

    def _update_planet(self, planet="", const_dir=None):
        """Add or update planetary constants."""
        self.planet = planet
        self.const_dir = const_dir
        self.const = init_const(self.planet, directory=self.const_dir)
        try:
            self.const.radius.convert_units("m")
            self._coord_system = GeogCS(semi_major_axis=self.const.radius.data)
        except AttributeError:
            self._coord_system = None
            _warn("Initialised without a coordinate system.")

    def _add_planet_conf_to_cubes(self):
        """Add or update planetary constants container to cube attributes."""
        add_planet_conf_to_cubes(self._cubes, self.const)

    def extract(self, constraints):
        """
        Subset `AtmoSim` using iris constraints.

        Parameters
        ----------
        constraints: iris.Constraint or iterable of constraints
            A single constraint or an iterable.
        """
        new_obj = self.__class__(
            cubes=self._cubes.extract(constraints),
            name=self.name,
            description=self.description,
            planet=self.planet,
            const_dir=self.const_dir,
            model=self.model,
            model_type=self.model_type,
            timestep=self.timestep,
            vert_coord=self.vert_coord,
        )
        return new_obj


class AtmoSim(AtmoSimBase):
    """
    Main class for dealing with a atmospheric model simulation output.

    Used to store and calculate atmospheric fields from gridded model output.
    Derived quantities are stored as cached properties to save time.
    """

    @cached_property
    @copy_doc(diag.sigma_p)
    def sigma_p(self):
        # TODO: pass the cube only?
        return diag.sigma_p(self._cubes, model=self.model)

    @cached_property
    @copy_doc(diag.wind_speed)
    def wind_speed(self):
        cmpnts = []
        for cmpnt_key in ["u", "v", "w"]:
            try:
                cmpnts.append(self[cmpnt_key])
            except AttributeError:
                pass
        return diag.wind_speed(*cmpnts)

    @cached_property
    @copy_doc(diag.toa_net_energy)
    def toa_net_energy(self):
        return diag.toa_net_energy(self._cubes, model=self.model)

    @cached_property
    @copy_doc(diag.sfc_water_balance)
    def sfc_water_balance(self):
        return diag.sfc_water_balance(
            self._cubes, const=self.const, model=self.model
        )



"""Functions for post-processing UM output."""

from typing import Optional

import iris
from iris.cube import CubeList

from .coord import (
    add_planet_calendar,
    ensure_bounds,
    interp_to_cube_time,
    regrid_3d,
    replace_z_coord,
    roll_cube_pm180,
)
from .exceptions import _warn
from .model import um
from .model.base import Model
from .subset import CM_INST_CONSTR, CM_MEAN_CONSTR, DimConstr, unique_cubes

__all__ = ("process_cubes",)


def process_cubes(
    cubelist: CubeList,
    timestep: Optional[int, float] = 1,
    ref_cube_constr: Optional[str] = um.thta,
    extract_incr: Optional[bool] = True,
    extract_mean: Optional[bool] = True,
    regrid_multi_lev: Optional[bool] = True,
    roll_pm180: Optional[bool] = True,
    add_calendar: Optional[bool] = False,
    calendar: Optional[dict] = None,
    planet: Optional[str] = "earth",
    remove_duplicates: Optional[bool] = True,
    use_varpack: Optional[bool] = False,
    varpack: Optional[dict] = None,
    interp_time: Optional[bool] = True,
    model: Optional[Model] = um,
) -> CubeList:
    """Post-process data for easier analysis."""
    if varpack is None:
        varpack = {}
    if calendar is None:
        calendar = {}
    DC = DimConstr(model=model)

    if remove_duplicates:
        cubelist = unique_cubes(cubelist)

    cm_constr = iris.Constraint()
    if extract_mean:
        cm_constr &= CM_MEAN_CONSTR
    else:
        cm_constr &= CM_INST_CONSTR
    cubelist = cubelist.extract(cm_constr)

    cubes = CubeList()

    # First, extract all multi-level fields
    if use_varpack:
        cubes += cubelist.extract(varpack["multi_level"])

        # Increments
        if extract_incr:
            cubes += cubelist.extract(
                iris.AttributeConstraint(
                    STASH=lambda x: x.item in [181, 182, 233]
                    and x.section not in [0]
                )
            )
    else:
        cubes = cubelist.extract(DC.strict.tmyx)

    if regrid_multi_lev:
        # Interpolation & regridding to common grid
        ref_cube = cubes.extract_cube(ref_cube_constr)
        ref_cube = replace_z_coord(ref_cube, model=model)

        # Interpolate to common levels
        cubes = CubeList(
            [
                regrid_3d(
                    replace_z_coord(cube, model=model), ref_cube, model=model
                )
                for cube in cubes
            ]
        )
    # Fix units of increments
    for cube in cubes:
        if cube.attributes["STASH"].item in [181, 233]:
            incr_unit = "K"
        elif cube.attributes["STASH"].item in [182]:
            incr_unit = "kg kg^-1"
        if cube.attributes["STASH"].item == 233:
            cube.units = f"{incr_unit} s^-1"
        elif cube.attributes["STASH"].item in [181, 182]:
            cube.units = f"{1/timestep} {incr_unit} s^-1"
            cube.convert_units(f"{incr_unit} s^-1")

    # Add all single-level cubes
    if use_varpack:
        cubes += cubelist.extract(varpack["single_level"])
    else:
        cubes += cubelist.extract(DC.strict.tyx)

    # Roll cubes to +/- 180 degrees in longitude for easier analysis
    if roll_pm180:
        rolled_cubes = CubeList()
        for cube in cubes:
            r_c = roll_cube_pm180(cube)
            ensure_bounds(r_c)
            rolled_cubes.append(r_c)
    else:
        rolled_cubes = cubes

    if interp_time:
        ref_cube = rolled_cubes.extract_cube(model.t_sfc)  # TODO: make an arg
        final = CubeList()
        for cube in rolled_cubes:
            final.append(interp_to_cube_time(cube, ref_cube, model=model))
    else:
        final = rolled_cubes

    if add_calendar:
        try:
            cal = calendar[planet]
            for cube in final:
                add_planet_calendar(
                    cube,
                    model.t,
                    days_in_year=cal["year"],
                    days_in_month=cal["month"],
                    days_in_day=cal["day"],
                    planet=planet,
                )
        except KeyError:
            _warn(f"Calendar for {planet=} not found.")
    return final



"""Test the plot submodule."""

import iris.coords
import iris.cube
import pytest

from aeolus import plot


@pytest.fixture(scope="module")
def example_cube_2d():
    xc = iris.coords.DimCoord(
        [-1, 2, 3], units="degrees", standard_name="longitude"
    )
    yc = iris.coords.DimCoord(
        [10, 30, 50, 70], units="degrees", standard_name="latitude"
    )
    arr = [[0, 1, 2], [3, 4, 5], [6, 7, 8], [9, 10, 11]]
    cube = iris.cube.Cube(
        data=arr,
        dim_coords_and_dims=[i[::-1] for i in [*enumerate((yc, xc))]],
        standard_name="x_wind",
        units="m/s",
    )
    return cube


def test_cube_minmeanmax_str(example_cube_2d):
    assert (
        plot.cube_minmeanmax_str(example_cube_2d) == "min=0 | mean=4 | max=11"
    )


def test_unit_format():
    assert plot.unit_format(-1.234e-5) == "$-1.2\\times10^{-5}$"
    assert plot.unit_format(6) == "$6.0$"
    assert (
        plot.unit_format(1.234e5, unit="m s^{-1}")
        == "$1.2\\times10^{5}$ $m$ $s^{-1}$"
    )
    assert plot.unit_format(1.234, decimal_digits=1, precision=3) == "$1.200$"
    assert plot.unit_format(1.234, decimal_digits=3, precision=1) == "$1.2$"
